// Crystann - GPL License - Xavier "critor" Andréani - TI-Planet.org

#include <math.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "CE.h"


#define min(a,b)            (((a)<(b))?(a):(b))
#define max(a,b)            (((a)>(b))?(a):(b))
#define dist(a,b)         (max(a,b)-min(a,b))
#define rgb2color(r, g, b)  ((uint16_t)(((r << 8) & 0xf800) | ((g << 3) & 0x07e0) | (b >> 3)))
#define printRess()         printSmall8bpp(uitoa(pers[0].ress,  textbuffer, 3), 6*TILE_SIZE+8, ystatus1)
#define printDiam()         printSmall8bpp(uitoa(pers[0].diam,  textbuffer, 3), 6*TILE_SIZE+8, ystatus2)
#define printScore()        printSmall8bpp(uitoa(score,  textbuffer, 4), 50, ystatus2)
#define printLife()         printSmall8bpp(uitoa(pers[0].force,  textbuffer, 3), 50, ystatus1)


/* Put your structures here */

typedef struct {
    uint8_t x;
    uint8_t xold;
    uint8_t y;
    uint8_t yold;
    uint8_t rx;
    uint8_t ry;
    uint8_t ress;
    uint8_t diam;
    uint8_t jumps;
    uint8_t force;
    uint8_t status;
    uint8_t idtile;
    uint8_t idtileold;    
    uint8_t idtilep;
} pers_t;


/* Put all (ALL except main) the functions prototypes here */

void printSmall8bpp(const char* string, uint16_t x, uint8_t y);
void cleanUp();

void reverse(char* s);
char* uitoa(uint24_t n, char* s, uint8_t zeropad_len);

void gc_drawTile(uint8_t* data, uint16_t x, uint16_t y, uint8_t w, uint8_t h);
void gc_drawTileTrsp(uint8_t* data, uint16_t x, uint16_t y, uint8_t w, uint8_t h);
uint8_t* decompTile(uint8_t* data_out, uint8_t* data_in);
uint8_t* decompMap(uint8_t* data_out, uint8_t* data_in, uint8_t retmode);
void setAutoCanvas();
void initPalette();
void* myMalloc(size_t size);
void myFreeLast();
void initTiles1(uint8_t** tiles);
void initTiles2(uint8_t** tiles);
void initTilesP(uint8_t** tilesp);
uint8_t* initMap();
uint8_t* initEndRoom();
void nopwait(uint16_t n);
uint8_t isKeyPressed(uint16_t key);
void scanKeys();
void addCleared(uint8_t x, uint8_t y);
void delayCleared(uint8_t x, uint8_t y);
uint16_t updCleared(uint16_t i);
void setTile(uint8_t* ptr, uint8_t val);
void resetP(uint8_t i, uint8_t x, uint8_t y);
void lightRoom(uint8_t x, uint8_t y);
void updSpirale();
uint8_t canGo(uint8_t p, uint8_t t, uint8_t x, uint8_t y);
void move(uint8_t p, uint8_t x0, uint8_t y0);
void myMain();

#define canHang(p,t)        (t==6 && !p || t==5 || t==16)
#define encodeCoords(x, y)  ((((uint32_t)(y))<<16)|(x))
#define getMapTilePtr(x, y) (map+(y)*MAP_WIDTH+(x))
#define getTile(val)        ((val)&0x7F)


/* Put all your globals etc. below */

#define MAXLIFE     100
#define X_START     38
#define Y_START     20
#define WAIT_DELAY1 5000
#define WAIT_DELAY2 2000
#define CLEAR_DELAY 31
#define MAXPERS     0x3
#define NTILES      61
#define NTILESP     21
#define SCR_WIDTH   320
#define SCR_HEIGHT  240
#define ROOM_WIDTH  15
#define ROOM_HEIGHT 12
#define BACKCOL_ID  6
#define MAP_WIDTH   150
#define MAP_HEIGHT  72
#define MAP_SIZE    (MAP_WIDTH*MAP_HEIGHT)
#define MAP_RWIDTH  (MAP_WIDTH/ROOM_WIDTH)
#define MAP_RHEIGHT  (MAP_HEIGHT/ROOM_HEIGHT)
#define TILE_SIZE   16
#define SCR_COLS    SCR_WIDTH/16
#define SCR_LINES   (SCR_HEIGHT/16-1)
#define STXT_HEIGHT 8
#define IS_FALLING  1
#define HASFIOLE    4
#define HASBOUEE    8
#define HASLAMPE    16
#define HASCLEF     32
#define HASPARCH    64
#define HASSTATUE   128

// key format : 2 bytes : keys group last addr byte + read mask
#define KEY_2ND     0x1220
#define KEY_MODE    0x1240
#define KEY_DEL     0x1280
#define KEY_ALPHA   0x1480
#define KEY_PLUS    0x1C02
#define KEY_MINUS   0x1C04
#define KEY_CLEAR   0x1C40
#define KEY_DOWN    0x1E01
#define KEY_LEFT    0x1E02
#define KEY_RIGHT   0x1E04
#define KEY_UP      0x1E08

// Global non-pointers variables (DON'T FORGET TO INITIALIZE THEM)
uint16_t ncleared  = 0;
uint16_t score = 0;
uint16_t xold=MAP_WIDTH, yold=0;
uint16_t x=0, y=0;
uint8_t ystatus1 = SCR_HEIGHT-2*STXT_HEIGHT, ystatus2 = SCR_HEIGHT-STXT_HEIGHT, ystatus3=SCR_HEIGHT-3*STXT_HEIGHT/2;
uint8_t sx=0, sy=0;

uint8_t* brightness = (uint8_t*)0xF60024;
uint8_t minBrightness = 229; // Do NOT set higher, may damage the hardware.
uint8_t maxBrightness = 49; // Do NOT set lower, may damage the hardware.
uint8_t userBrightness = 100; // will get overwritten

// Global arrays (DON'T FORGET TO INITIALIZE THEM)
uint8_t reachable[NTILES]   = {1,0,0,0,0,1,1,0 ,0 ,0 ,1 ,1 ,0 ,1 ,1 ,0 ,1 ,1 ,1 ,0 ,0 ,0 ,1 ,1 ,0 ,0 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,0 ,0,0 ,0,0 ,0 ,0, 1, 1, 1, 1, 1, 1,0 ,0 ,0 ,0 ,0,0 ,0 ,0 ,0 ,0};
uint8_t nextanim[NTILES]    = {0,1,2,3,4,5,6,37,39,41,10,11,12,36,14,15,16,44,18,19,20,21,22,23,24,25,46,48,28,29,30,31,32,33,34,35,13,38,7,40,8,42,43,9,45,17,47,26,49,27,51,52,53,54,0,55,56,57,58,59};
//pers_t pers[MAXPERS] = {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}};
pers_t* pers = {0};

uint32_t* cleared    = {0};
uint8_t* spirale     = {0};
uint8_t* light       = {0};
uint8_t* screentiles = {0};
uint8_t* map         = {0};
char textbuffer[5]   = {0};

// Pointers
uint16_t* palette = (uint16_t*)0xE30200;
static const uint8_t* screen_buf = (uint8_t*)0xD40000;
static const uint8_t* offscreen_ptr = (uint8_t*)0xD52C00;    // vram "back buffer" (free to use when LCD is in 8bpp)  76.8k
static const uint8_t* safeRAM_orig_ptr = (uint8_t*)0xD07396; // 52.290k
uint8_t* safeRAM_ptr = (uint8_t*)0xD07396; // safeRAM_orig_ptr
uint16_t safeRAM_free = 52290; // D07396h: cmdPixelShadow (8400 b) + D09466h: plotSScreen (21945 b) + D0EA1Fh: saveSScreen (21945 b)
uint8_t* safeRAM_ptr_old = (uint8_t*)0xD07396;


/* Put all your code here, with main() first */

void main()
{
    asm("PUSH IY");
    asm("CALL _RunIndicOff");
    myMain();
    asm("POP IY");
    cleanUp();
}

void myMain()
{
    uint16_t i=0, i2=0, j=0, j2=0;
    uint8_t t=0, tbot=0, t2=0, t3=0, k=0, x0=0, y0=0, k2=0;
    uint8_t *pct=0, *pcr=0;
    int key = 0;
    uint8_t* tiles[NTILES]   = {0};
    uint8_t* tilesp[NTILESP] = {0};
    pers=(pers_t*)myMalloc(MAXPERS*sizeof(pers_t));

    asm("di");
    asm("xor a,a");
    asm("ld (kbdGetKy),a"); // flush keys

    userBrightness = *brightness;

    asm("CALL _ClrScrn");

    // 320x240 8bpp
    initPalette();
    memset(screen_buf, 0, SCR_HEIGHT*SCR_WIDTH*2);
    *brightness = minBrightness;
    asm("LD A,27h");
    asm("LD (E30018h),A");
    memset(screen_buf, 0, SCR_HEIGHT*SCR_WIDTH);
    *brightness = userBrightness;
 
    map = initMap();
    screentiles = myMalloc(SCR_COLS*SCR_LINES);
    initTilesP(tilesp);
    initTiles1(tiles);
    initTiles2(tiles);

    memset(screen_buf+(SCR_HEIGHT-TILE_SIZE)*SCR_WIDTH, 17, TILE_SIZE*SCR_WIDTH);
    cleared = myMalloc(SCR_COLS*SCR_LINES*4);
    spirale = myMalloc(MAP_RWIDTH*MAP_RHEIGHT);
    light = myMalloc(MAP_RWIDTH*MAP_RHEIGHT);
    memset(spirale,0xFF,MAP_RWIDTH*MAP_RHEIGHT);
    memset(light,0,MAP_RWIDTH*MAP_RHEIGHT);
    pct=map;
    
    for(j=0;j<MAP_HEIGHT;j++)
    {
        for(i=0;i<MAP_WIDTH;i++)
        {
            if(getTile(*pct)==13)
                spirale[j/ROOM_HEIGHT*MAP_RWIDTH+i/ROOM_WIDTH]=((j%ROOM_HEIGHT)<<4) | (i%ROOM_WIDTH);
            pct++;            
        }
    }    
    printSmall8bpp("Force :",1,ystatus1);
    printSmall8bpp("Score :",1, ystatus2);
    gc_drawTileTrsp(tiles[11],5*TILE_SIZE+1,SCR_LINES*TILE_SIZE,TILE_SIZE,TILE_SIZE);
    printSmall8bpp("x",6*TILE_SIZE-1,ystatus1);
    gc_drawTileTrsp(tiles[10],8*TILE_SIZE+7,SCR_LINES*TILE_SIZE,TILE_SIZE,TILE_SIZE);
    printSmall8bpp("x",8*TILE_SIZE+1,ystatus2);
    printSmall8bpp("[2nd]+[<>",11*TILE_SIZE+6,ystatus1);
    printSmall8bpp("] clear wall",15*TILE_SIZE,ystatus1);
    printSmall8bpp("[2nd]+[ ^",11*TILE_SIZE+6,ystatus2);    
    printSmall8bpp("] big jump",15*TILE_SIZE,ystatus2);    
    pers[0].status=0;
    pers[0].ress=0;
    pers[0].diam=0;
    pers[0].x=0;
    pers[0].y=0;
    for(i=0;i<MAXPERS;i++)
    {
        resetP(i,X_START,(!i)?Y_START:MAP_HEIGHT);
        if(i>0)
            pers[i].force=0;
    }
    printRess();
    printDiam();
    lightRoom(pers[0].rx,pers[0].ry);
    setAutoCanvas();    
    asm("ei");

    scanKeys();
    while (!isKeyPressed(KEY_MODE))
    {
       pcr=screentiles;
       pct = getMapTilePtr(x,y);
       for (j=0,j2=y;j<SCR_LINES;j++,j2++)
       {   
           for (i=0,i2=x;i<SCR_COLS;i++,i2++)
           {
               t = *pct;
               if(t&0x80) 
                   t=NTILES-1;
               else
                   *pct=nextanim[getTile(*pct)];
               t2=1;
               t3=0;
               for(k=0;k<MAXPERS && t2;k++)
                   if (pers[k].x==i2 && pers[k].y==j2)
                   {
                       gc_drawTile(tilesp[pers[k].idtilep], i*TILE_SIZE, j*TILE_SIZE, TILE_SIZE, TILE_SIZE);
                       t2=0;
                   }
                   else if(pers[k].xold+(x-xold)==i2 && pers[k].yold+(y-yold)==j2 && (i2!=pers[k].x || j2!=pers[k].y))
                       t3=1;
               if(t2 && (xold>=MAP_WIDTH || *pcr!=t || t3))
               {
                   gc_drawTile(tiles[t], i*TILE_SIZE, j*TILE_SIZE, TILE_SIZE, TILE_SIZE);
                   *pcr = t;
               }
               pct++;
               pcr++;
           }
           pct+=MAP_WIDTH-SCR_COLS;
       }
       for (k=0;k<ncleared;k++)
       {
            if (!updCleared(k))
            {
                i=cleared[k]&0xFF;
                j=cleared[k]>>8;
                t2=1;
                for(t=0;t<MAXPERS && t2;t++)
                    if(pers[t].force && pers[t].x==i && pers[t].y==j)
                        t2=0;
                if(t2)
                {
                    t=map[j*MAP_WIDTH+i];
                    if(!t) t=55;
                    else if(t>=55 && t<=58) t++;
                    else
                    {
                        t=12;
                        if(k==ncleared-1)
                            ncleared--;
                    }
                    map[j*MAP_WIDTH+i]=t;
                }
            }
       }
       xold = x;
       yold = y;
       x0=pers[0].xold;
       y0=pers[0].yold;        
       for(k=MAXPERS-1;k<MAXPERS;k--)
       {
           pers[k].idtileold=pers[k].idtile;
           if(!k && pers[0].jumps && pers[0].xold==pers[0].x && pers[0].yold==pers[0].y)
                   pers[0].jumps--;
           pers[k].xold=pers[k].x;
           pers[k].yold=pers[k].y;
           if(!(pers[k].status&IS_FALLING) && !canHang(k,pers[k].idtile) && !pers[k].jumps)
           {
               t=(pers[k].y<MAP_HEIGHT-1)?getTile(*(getMapTilePtr(pers[k].x, pers[k].y+1))):1;
               if(canGo(k,t,pers[k].x, pers[k].y+1) && t!=5)
                   pers[k].status|=IS_FALLING;
           }
           move(k,x0,y0);
           if(pers[k].y>=MAP_HEIGHT)
           {    
               if(!k)
               {
                   pers[0].y=pers[0].yold;
                   if(score>MAXLIFE)
                       score-=MAXLIFE;
                   else
                       score=0;
                   printScore();
                   resetP(0,X_START,Y_START);
               }
               else if(sy<MAP_HEIGHT && pers[0].force && pers[0].xold/ROOM_WIDTH==pers[0].rx && pers[0].yold/ROOM_HEIGHT==pers[0].ry && canGo(t,13,sx,sy))
                   resetP(k,sx,sy);
           }
           else if (pers[k].force && (pers[k].x!=pers[k].xold || pers[k].y!=pers[k].yold))
           {
               if(!k)
               {
                   t=getTile(*getMapTilePtr(pers[0].x,pers[0].y));
                   if(t==9)
                   {
                       pers[0].force=100;
                       printLife();
                   }
                   else if(pers[0].status&HASSTATUE && (t==19 || t==20 || t==24 || t==25))
                       initEndRoom();
                   if(isKeyPressed(KEY_2ND) && pers[0].x!=pers[0].xold && !pers[0].jumps)
                   {
                        pct=getMapTilePtr(pers[0].x,pers[0].y+1);
                        if(getTile(*pct)==12 && pers[0].status&HASFIOLE)
                        {
                            *pct=50;
                            addCleared(pers[0].x,pers[0].y+1);
                            pers[0].x=pers[0].xold;
                            pers[0].y=pers[0].yold;
                        }
                   }
               }
               else if(pers[0].force && pers[0].x==pers[k].x && pers[0].y==pers[k].y)
               {
                   pers[0].force--;
                   printLife();
               }
               if(pers[k].x!=pers[k].xold)
               {
                   pct=getMapTilePtr(pers[k].x,pers[k].yold);
                   if(!canGo(k,getTile(*pct),pers[k].x,pers[k].yold))
                   {
                       pers[k].x=pers[k].xold;
                   }
                   else
                       pers[k].y=pers[k].yold;
               }
               if(!k && isKeyPressed(KEY_2ND) && pers[0].y<pers[0].yold && pers[0].ress && !pers[0].jumps)
               {
                    pct=getMapTilePtr(pers[0].x,pers[0].y);
                    t=getTile(*getMapTilePtr(pers[0].x,pers[0].yold+1));
                    if(canGo(0,getTile(*pct),pers[0].x,pers[0].y) && (canHang(0,pers[0].idtile) || !canGo(0,t,pers[0].x,pers[0].yold+1) || t==5))
                    {
                       pers[0].jumps=5;
                       pers[0].ress--;
                       printRess();
                       score-=5;
                       printScore();
                    }
                    else
                       pers[0].y=pers[0].yold;
               }
               else if(pers[k].y!=pers[k].yold)
               {
                   if(pers[k].y>pers[k].yold)
                       pers[k].jumps=0;
                   pct=getMapTilePtr(pers[k].x,pers[k].y);
                   if(!canGo(k,getTile(*pct),pers[k].x,pers[k].y) || pers[k].y<pers[k].yold &&  pers[k].idtile!=5 && !pers[k].ress)
                   {
                       pers[k].y=pers[k].yold;
                   }
               }
               if (pers[k].x==pers[k].xold && pers[k].y==pers[k].yold)
               {
                   pers[k].status=pers[k].status&~IS_FALLING;
               }
               else
               {
                   pers[k].rx=pers[k].x/ROOM_WIDTH;
                   pers[k].ry=pers[k].y/ROOM_HEIGHT;
                   if(!k)
                       updSpirale();
                   if(pers[k].jumps)
                       pers[k].jumps--;
                   t=getTile(*pct);
                   if((*pct)&0x80 || !k && !light[pers[0].ry*MAP_RWIDTH+pers[0].rx]) {
                       if(!k && (pers[0].status&HASLAMPE || pers[0].rx!=7 || pers[0].ry!=2))
                           lightRoom(pers[0].rx,pers[0].ry);
                       else
                           *pct=t;
                   }
                   pers[k].idtile=t;
                   if(canHang(k,t) && (t!=16 || pers[k].idtileold==16))
                       pers[k].status=pers[k].status&~IS_FALLING;
                   if(!k && (t==11 || t==10 || t==18 || (t>=21 && t<=23) || (t>=28 && t<=30)))
                   {
                       if(t==11)
                       {
                           pers[0].ress+=2;
                           printRess();
                       }
                       else if(t==10)
                       {
                           pers[0].diam++;
                           if(pers[0].diam==105)
                               *getMapTilePtr(2*ROOM_WIDTH+7,ROOM_HEIGHT+4)=30;
                           printDiam();
                       }
                       else if(t==18)
                       {
                           pers[0].status|=HASFIOLE;
                           gc_drawTileTrsp(tiles[t],9*TILE_SIZE+4,SCR_LINES*TILE_SIZE,TILE_SIZE,TILE_SIZE);
                       }
                       else if(t==22)
                       {
                           pers[0].status|=HASLAMPE;   
                           gc_drawTileTrsp(tiles[t],12*TILE_SIZE+4,SCR_LINES*TILE_SIZE,TILE_SIZE,TILE_SIZE);
                       }                   
                       else if(t==23)
                       {
                           pers[0].status|=HASBOUEE;   
                           gc_drawTileTrsp(tiles[t],10*TILE_SIZE+4,SCR_LINES*TILE_SIZE,TILE_SIZE,TILE_SIZE);
                       }
                       else if(t==28)
                       {
                           pers[0].status|=HASCLEF;   
                           gc_drawTileTrsp(tiles[t],11*TILE_SIZE+4,SCR_LINES*TILE_SIZE,TILE_SIZE,TILE_SIZE);
                       }
                       else if(t==29)
                       {
                           pers[0].status|=HASSTATUE;   
                           gc_drawTileTrsp(tiles[t],14*TILE_SIZE+4,SCR_LINES*TILE_SIZE,TILE_SIZE,TILE_SIZE);
                       }
                       else if(t==30)
                       {
                           pers[0].status|=HASPARCH;   
                           gc_drawTileTrsp(tiles[t],13*TILE_SIZE+4,SCR_LINES*TILE_SIZE,TILE_SIZE,TILE_SIZE);
                           *getMapTilePtr(9*ROOM_WIDTH,3*ROOM_HEIGHT+3)=0;
                       }
                       score+=20;
                       printScore();
                       *pct=0;
                   }
               }
           }
           t=pers[k].idtile;
           tbot=(pers[k].y<MAP_HEIGHT-1)?getTile(*(getMapTilePtr(pers[k].x, pers[k].y+1))):1;
           if(pers[k].force && (t==16 && !(pers[k].status&HASBOUEE) || ((tbot==15 || tbot==7 || tbot>=37 && tbot<=40 || tbot==8 || tbot==16 && !(pers[k].status&HASBOUEE)) && !pers[k].jumps && !canHang(k,t))))
           {
               pers[k].force=0;
               if(!k)
                   printLife();
           }
           if(!k)
           {
               if(!pers[0].force)
                   pers[0].idtilep=(pers[0].idtilep!=16)?16:17;
               else if(pers[0].x<pers[0].xold && t!=5 && t!=16)
               {   
                   if(t==6)
                       pers[0].idtilep=(pers[0].idtilep!=9)?9:10;
                   else
                       pers[0].idtilep=(pers[0].idtilep!=2)?2:3;
               }
               else if(pers[0].x>pers[0].xold && t!=5 && t!=16)
               {   
                   if(t==6)
                       pers[0].idtilep=(pers[0].idtilep!=11)?11:12;
                   else
                       pers[0].idtilep=(pers[0].idtilep!=4)?4:5;
               }
               else if(t==6 && (pers[0].idtilep<8 || pers[0].idtilep>12))
                   pers[0].idtilep=8;
               else if(t==5 && (pers[0].x!=pers[0].xold || pers[0].y!=pers[0].yold))
                   pers[0].idtilep=(pers[0].idtilep!=6)?6:7;
               else if(t==16)
                   pers[0].idtilep=(pers[0].idtilep!=13)?13:14;
               else if(pers[0].status&IS_FALLING && t!=5 && t!=6 && t!=16)
                   pers[0].idtilep=15;
               else if(t!=6 && t!=5 && t!=16)
                   pers[0].idtilep=(pers[0].idtilep)?0:1;
           }
           else
           {
               if(pers[k].idtilep==18 || pers[k].idtilep==19)
                   pers[k].idtilep++;
               else
                   pers[k].idtilep=18;
           }
       }
       setAutoCanvas();

       nopwait(WAIT_DELAY1);
       if(x==xold && y==yold)
           nopwait(WAIT_DELAY2);
    }
}

void move(uint8_t p, uint8_t x0, uint8_t y0)
{
    uint8_t y=pers[0].y;
    if(!pers[p].force)
    {
        if(!p)
            pers[0].y--;
        else
            pers[p].y=MAP_HEIGHT;
    }
    else if(pers[p].status&IS_FALLING)
            pers[p].y++;
    else if(!p)
    {
        pers[0].x = pers[0].x + isKeyPressed(KEY_RIGHT) - isKeyPressed(KEY_LEFT);
        pers[0].y = pers[0].y + isKeyPressed(KEY_DOWN) - isKeyPressed(KEY_UP);
    }
    else if(sy<MAP_HEIGHT && (pers[0].rx!=pers[p].rx || pers[0].ry!=pers[p].ry) && dist(sx,pers[0].xold)+dist(sy,pers[0].yold)<dist(pers[p].x,pers[0].xold)+dist(pers[p].y,pers[0].yold))
    {   
        if(canGo(p,13,sx,sy) && pers[0].force && pers[0].xold/ROOM_WIDTH==pers[0].rx && pers[0].yold/ROOM_HEIGHT==pers[0].ry)
            pers[p].force=0;
    }
    else
    {
        pers[0].y=MAP_HEIGHT;
        if(pers[p].idtile==5 && pers[p].y>y0 && canGo(p,*getMapTilePtr(pers[p].x,pers[p].y-1),pers[p].x,pers[p].y-1))
            pers[p].y--;
        else if(pers[p].y<y0 && canGo(p,*getMapTilePtr(pers[p].x,pers[p].y+1),pers[p].x,pers[p].y+1))
            pers[p].y++;
        else if(pers[p].x>x0 && canGo(p,*getMapTilePtr(pers[p].x-1,pers[p].y),pers[p].x-1,pers[p].y))
            pers[p].x--;
        else if(pers[p].x<x0 && canGo(p,*getMapTilePtr(pers[p].x+1,pers[p].y),pers[p].x+1,pers[p].y))
            pers[p].x++;
        else if(pers[p].y>y0)
            pers[p].y--;    
        pers[0].y=y;
    }
}

uint8_t canGo(uint8_t p, uint8_t t, uint8_t x, uint8_t y)
{
    uint8_t r=(((t==21) && (pers[p].status&HASCLEF)) || reachable[t]);
    uint8_t k=0;
    for(k=0;r && k<MAXPERS;k++)
    {
        if(k!=p && pers[k].force && pers[k].x==x && pers[k].y==y)
            r=0;
    }
    return r;
}

void updSpirale()
{
    uint8_t t=spirale[pers[0].ry*MAP_RWIDTH+pers[0].rx];
    sx=t&0xF;
    sy=t>>4;
    if(sy>=ROOM_HEIGHT)
        sy=MAP_HEIGHT;
    else
    {
        sx+=pers[0].rx*ROOM_WIDTH;
        sy+=pers[0].ry*ROOM_HEIGHT;
    }
}

void lightRoom(uint8_t x,uint8_t y)
{
    uint8_t x0=x*ROOM_WIDTH;
    uint8_t y0=y*ROOM_HEIGHT;
    uint8_t i=0,j=0;
    uint8_t* ptr=0;
    ptr=getMapTilePtr(x0,y0);
    for(j=0;j<ROOM_HEIGHT;j++)
    {   
        for(i=0;i<ROOM_WIDTH;i++)
        {
             *ptr&=~0x80;
             ptr++;
        }
        ptr+=MAP_WIDTH-ROOM_WIDTH;
    }
    light[y*MAP_RWIDTH+x]=1;
    score+=20;
    printScore();
}

void resetP(uint8_t i, uint8_t x, uint8_t y)
{
    pers[i].xold=pers[i].x;
    if(pers[i].y<MAP_HEIGHT)
        pers[i].yold=pers[i].y;
    pers[i].x=x;
    pers[i].y=y;
    pers[i].rx=x/ROOM_WIDTH;
    pers[i].ry=y/ROOM_HEIGHT;
    pers[i].idtilep=(!i)?0:18;
    pers[i].idtile = (y<MAP_HEIGHT)?getTile(*(getMapTilePtr(pers[i].x, pers[i].y))):0;
    pers[i].jumps=0;
    pers[i].force=MAXLIFE;
    pers[i].status&=~IS_FALLING;
    if(i)
        pers[i].ress=1;
    else
    {
        updSpirale();
        printLife();
    }
}

void setTile(uint8_t* ptr, uint8_t val)
{
    *ptr=((*ptr)&0x80)|val;   
}

void addCleared(uint8_t x, uint8_t y)
{
    uint32_t val = (CLEAR_DELAY<<15) | (((uint32_t)y)<<8) | ((uint32_t)x);
    uint8_t i=0;
    for (i=ncleared;i>0;i--)
        cleared[i]=cleared[i-1];
    ncleared++;
    cleared[0]=val;
}

void delayCleared(uint8_t x, uint8_t y)
{
    uint32_t val = (0<<15) | (((uint32_t)y)<<8) | ((uint32_t)x);
    uint8_t i=0;
    for (i=ncleared;i>0;i--)
        cleared[i]=cleared[i-1];
    ncleared++;
    cleared[0]=val;
}


uint16_t updCleared(uint16_t i)
{
    uint32_t n = cleared[i]>>15;
    if (n>0) 
    {   
        n--;
        cleared[i] = (cleared[i]&0x7FFF) | (n<<15);
    }
    return n;
}

void nopwait(uint16_t n)
{
    while(n)
    {
        asm("nop"); // small delay
        n--;
    }
}

uint8_t* decompMap(uint8_t* data_out, uint8_t* data_in, uint8_t retmode)
{
    uint8_t n, id;
    while (*data_in)
    {
        n = (*data_in)>>5;
        id = ((*data_in)%32)|0x80;
        memset(data_out, id, n);
        data_out += n;
        data_in++;
    }
    return retmode?(data_in+1):data_out;
}

uint8_t* initEndRoom()
{
    uint8_t map_C[] = {
                        225,225,33,0,
                        225,225,33,0,
                        161,34,64,68,35,129,0,
                        97,34,36,192,36,97,0,
                        65,34,224,64,35,65,0,
                        65,224,128,65,0,
                        65,39,224,64,40,65,0,
                        65,224,128,65,0,
                        65,224,128,65,0,
                        65,70,225,129,0,
                        65,64,129,191,65,0,
                        97,47,225,129,0,
                      };
    uint8_t* ptr_in = map_C, i=0;
    for(i=0; i<ROOM_HEIGHT; i++)
    {
       ptr_in = decompMap(getMapTilePtr(pers[0].rx*ROOM_WIDTH,pers[0].ry*ROOM_HEIGHT+i), ptr_in, 1);
    }
    ptr_in=getMapTilePtr(pers[0].rx*ROOM_WIDTH+9,pers[0].ry*ROOM_HEIGHT+10);
    for(i=0;i<4;i++)
    {
        *ptr_in=32+i;
        ptr_in++;
    }
}

uint8_t* initMap1(uint8_t* map)
{
    uint8_t map_C[] = {225,225,225,225,225,225,225,225,225,225,225,225,225,225,225,225,225,225,225,225,225,129,198,97,64,41,32,65,36,64,68,32,36,32,45,32,100,65,68,96,36,32,45,96,68,65,224,192,65,160,45,224,192,68,128,100,65,224,128,68,65,134,161,102,33,134,33,166,225,36,128,68,33,32,193,224,32,33,37,32,74,65,224,224,224,224,65,39,128,74,160,40,65,39,224,128,43,224,224,224,224,224,224,224,224,128,65,224,33,64,41,64,65,42,64,43,128,33,37,161,39,224,224,128,42,32,42,192,65,128,140,160,65,192,42,96,37,129,37,224,96,42,224,160,42,224,224,224,224,160,65,224,64,43,32,43,42,65,44,32,33,34,32,33,37,44,33,37,96,65,224,43,224,37,32,35,33,108,33,96,33,134,37,192,79,96,65,96,37,108,37,70,37,225,225,225,225,225,79,33,79,97,79,33,79,33,47,97,47,224,32,65,192,37,225,33,160,45,37,32,33,230,134,204,46,76,33,37,160,44,128,44,37,96,37,128,37,32,65,224,32,37,96,37,64,37,224,224,96,65,134,44,45,192,36,225,225,225,33,224,40,65,39,160,37,192,65,192,37,224,32,37,192,37,32,43,96,65,76,37,102,44,128,33,37,65,32,37,96,69,102,37,224,37,96,37,64,37,160,43,192,42,96,43,65,96,43,128,37,128,65,224,192,65,224,192,65,192,37,102,161,39,160,37,224,32,37,140,37,32,37,65,96,65,64,37,32,44,37,64,37,76,33,37,65,37,128,37,128,37,76,225,225,129,44,34,37,32,35,140,34,37,97,37,44,46,33,128,37,44,37,32,40,65,39,192,43,64,42,64,65,39,224,160,65,224,64,37,96,65,192,37,224,32,37,128,37,32,37,224,64,37,32,33,37,32,43,37,96,37,65,37,134,37,224,224,224,43,65,96,37,224,37,32,65,37,64,33,96,43,37,32,37,224,37,33,140,33,44,33,32,65,224,192,65,32,33,32,49,58,59,49,58,59,37,32,33,32,65,224,224,192,37,32,37,224,64,37,32,33,108,33,37,64,37,65,37,224,64,108,65,44,37,224,32,37,44,46,65,32,33,39,37,224,37,96,37,64,33,37,44,46,44,34,32,76,37,65,37,64,37,65,172,33,32,65,160,49,58,59,49,58,59,129,32,65,240,65,32,65,96,43,224,160,42,96,37,32,37,224,64,37,32,33,96,33,37,64,37,65,224,192,65,42,37,192,43,32,37,42,32,65,32,37,42,37,224,37,96,37,96,37,224,37,65,37,64,37,33,192,33,32,65,128,33,208,129,42,33,240,80,33,32,97,37,225,33,96,225,225,225,65,37,225,97,79,33,79,65,47,225,225,97,32,97,37,225,225,225,225,161,192,161,96,33,208,193,240,80,161,37,225,33,96,225,225,225,65,37,225,225,225,225,225,32,97,37,225,225,225,225,161,192,161,96,33,208,193,240,80,129,32,37,38,37,32,45,96,33,96,65,64,196,224,224,192,65,224,192,65,192,45,64,43,32,33,32,65,32,37,68,96,68,32,100,65,224,192,65,32,45,32,68,224,32,65,128,33,208,129,39,97,240,129,96,37,128,40,33,96,65,39,224,224,224,32,44,230,230,33,96,43,65,64,42,192,35,44,33,32,65,32,37,224,128,65,108,39,224,32,40,65,224,160,40,65,128,35,33,208,97,74,32,161,144,97,96,37,38,37,96,33,96,65,224,224,224,192,40,65,39,224,64,75,42,65,37,44,34,224,32,33,32,65,39,37,160,33,224,32,65,224,224,32,33,224,64,65,160,35,33,176,161,37,96,37,65,112,97,160,37,96,33,96,65,160,35,65,172,65,44,37,224,128,65,160,75,64,37,35,129,37,224,64,40,33,32,65,32,37,160,33,39,224,53,224,192,42,32,37,33,224,32,40,65,39,160,35,33,240,48,161,80,33,112,97,32,43,42,64,37,38,37,32,33,47,64,65,192,68,160,65,32,37,64,129,37,33,64,40,65,128,35,65,34,32,37,96,65,37,64,37,108,34,96,33,32,65,32,37,160,33,96,35,44,65,108,37,33,140,33,37,76,65,76,37,33,224,64,65,192,40,65,240,144,65,176,97,37,76,34,96,37,32,65,64,65,39,224,160,65,39,37,32,33,34,68,32,37,35,33,64,65,39,224,32,37,64,40,65,37,43,32,37,224,33,32,65,32,37,160,33,42,128,65,96,37,33,128,33,37,64,65,64,37,224,96,65,224,32,35,225,33,144,97,112,97,37,192,37,32,37,33,32,97,224,192,65,32,37,65,32,51,56,32,37,32,65,32,65,224,64,37,96,65,76,34,37,192,42,33,32,65,39,37,160,33,76,46,129,39,64,37,33,128,33,37,32,40,65,192,35,76,34,37,224,192,100,32,161,240,80,97,37,230,38,37,33,64,65,96,43,32,50,160,42,32,65,39,32,65,32,52,57,32,37,32,65,32,65,224,64,37,96,65,96,37,128,35,76,33,32,65,32,37,224,128,65,96,37,33,128,33,37,64,97,230,70,37,224,224,64,97,64,35,65,176,161,224,64,37,33,64,65,64,193,64,47,129,64,193,37,97,32,65,64,107,128,37,64,43,65,39,64,37,224,33,32,65,32,37,224,128,65,96,37,33,128,33,37,64,65,224,96,37,35,44,129,37,192,35,97,192,35,161,34,64,65,192,43,64,37,33,64,65,64,193,64,161,64,37,225,33,37,32,65,235,42,32,37,74,43,65,96,37,128,37,64,33,32,65,32,37,160,43,96,75,65,96,37,33,64,74,33,37,64,65,32,43,224,32,37,32,42,65,43,32,37,128,42,64,53,32,43,128,42,224,32,42,32,161,37,225,64,65,64,193,64,225,37,225,33,37,225,225,225,161,37,97,32,97,37,225,225,65,37,33,44,129,37,225,225,225,97,37,225,225,225,193,37,225,64,65,64,193,64,225,37,225,33,37,225,225,225,161,37,97,32,97,37,225,225,65,37,33,32,129,37,225,225,225,97,37,225,225,225,97,68,32,37,192,33,64,65,96,132,96,68,32,65,64,37,128,42,96,37,32,65,224,45,160,65,42,128,45,64,37,64,33,32,65,32,37,224,96,42,65,96,37,33,64,33,39,32,37,64,65,64,65,68,32,36,96,68,65,160,37,224,65,64,36,32,45,32,68,128,36,65,224,64,40,33,64,65,224,160,40,65,64,37,64,35,76,34,64,37,32,97,32,129,37,64,33,102,65,76,37,166,37,32,40,33,32,65,32,37,224,32,43,32,44,65,39,96,33,64,33,32,42,37,32,40,65,32,60,224,128,65,39,128,37,192,40,65,39,224,160,65,128,37,44,46,33,64,33,64,65,224,192,65,64,37,224,32,37,32,65,160,65,192,65,64,37,160,37,64,33,32,65,32,37,224,35,34,64,65,42,96,33,64,33,76,37,64,65,37,97,37,224,224,32,37,96,75,224,224,96,65,128,37,64,45,32,42,33,64,65,224,160,40,65,64,37,224,32,37,32,65,39,96,42,224,32,65,39,32,37,160,37,32,40,33,32,65,32,37,224,64,43,42,65,76,160,33,64,37,64,65,37,32,45,32,37,224,96,37,76,37,108,33,108,33,37,224,192,42,32,65,64,43,32,37,32,37,32,37,44,33,64,65,224,192,65,64,37,44,47,44,64,44,47,44,37,32,65,32,37,108,34,64,37,76,47,32,65,64,37,224,32,33,32,65,32,37,160,35,34,64,76,65,224,33,43,32,37,64,65,128,33,76,34,64,35,76,65,37,64,37,96,45,96,225,33,32,129,46,161,37,108,34,32,37,32,37,32,33,64,65,192,42,160,37,65,64,37,32,33,128,33,32,37,32,65,32,37,192,37,32,65,32,65,64,37,192,35,65,32,65,32,37,128,43,96,40,33,32,65,39,192,33,108,37,44,65,128,33,224,32,65,76,46,44,37,96,43,128,65,224,42,128,36,65,37,128,43,37,32,37,32,33,64,65,160,35,140,34,32,37,65,75,37,32,33,39,64,40,33,32,37,43,65,32,37,192,37,32,42,32,42,97,108,33,224,32,65,32,37,96,35,34,160,40,65,64,43,96,47,33,96,37,32,65,128,33,224,32,65,128,37,172,37,64,65,37,198,33,160,65,37,128,35,108,37,33,64,65,224,160,37,97,44,97,128,33,46,33,44,65,32,37,42,128,35,76,33,46,97,236,44,65,34,64,65,32,37,224,128,65,32,37,76,37,76,33,39,64,37,32,65,32,37,224,128,65,37,32,42,32,37,160,37,128,37,224,160,65,37,224,32,37,33,64,65,224,160,37,224,224,32,65,32,140,37,224,65,128,44,33,76,32,33,96,65,32,37,65,224,64,65,32,37,64,37,64,45,96,37,32,65,32,37,224,128,65,37,44,46,44,37,128,37,76,37,64,37,224,160,65,37,42,224,37,33,64,65,64,43,224,64,37,224,192,43,32,65,43,128,37,192,43,65,32,74,96,42,64,33,96,65,32,37,224,128,65,32,37,64,37,160,42,37,32,65,32,37,64,42,224,32,65,37,96,37,128,37,32,42,37,97,43,64,47,37,224,161,37,225,64,65,64,65,239,225,65,37,65,37,225,129,37,0};
    return decompMap(map, map_C, 0);
}

uint8_t* initMap2(uint8_t* map)
{
    uint8_t map_C[] = {225,65,140,193,79,129,37,65,239,225,225,161,47,37,33,47,65,239,225,225,225,33,37,225,161,37,225,64,65,64,225,225,129,37,65,37,225,129,37,225,65,128,225,161,37,225,225,225,225,33,37,225,225,225,225,161,37,225,65,96,37,64,45,96,33,64,65,64,36,128,37,32,100,32,65,32,45,96,37,64,37,128,65,160,37,224,65,224,192,65,32,37,160,45,160,65,224,192,65,32,37,96,45,224,65,64,100,32,132,32,68,65,96,33,32,37,224,65,96,37,160,43,33,64,65,64,43,32,54,32,33,37,128,40,65,192,97,108,37,65,160,37,70,37,128,65,39,224,160,65,32,37,224,43,42,64,65,128,43,96,42,128,65,32,37,224,96,40,65,39,224,160,65,96,33,32,37,192,40,65,39,32,35,44,34,37,64,76,33,64,65,64,35,46,97,37,160,65,192,97,96,37,65,39,128,37,64,37,96,40,65,224,161,32,65,32,37,108,46,204,37,65,96,37,33,102,33,102,44,65,32,37,224,128,65,224,64,43,128,33,96,33,32,37,224,65,160,37,128,33,64,65,224,192,65,76,34,96,65,128,37,65,128,69,64,69,96,65,224,33,32,97,32,65,32,37,224,96,37,65,96,37,224,160,37,224,44,64,40,65,39,192,37,225,65,37,33,32,37,38,37,160,65,160,37,134,33,64,65,96,49,58,59,49,58,59,49,58,59,32,65,192,65,39,96,37,65,128,37,128,37,96,65,39,192,33,64,41,33,32,65,32,37,224,96,37,65,39,64,37,224,160,37,64,33,44,64,76,96,65,224,37,160,65,64,37,128,37,160,65,39,128,37,128,33,64,65,64,33,240,80,97,192,65,128,37,65,39,64,69,128,69,32,40,65,224,53,96,33,32,65,32,37,32,42,224,32,37,65,96,37,224,76,65,76,33,32,76,46,32,42,32,33,64,65,39,192,37,128,40,65,64,37,32,79,32,37,96,129,96,42,32,37,128,33,64,65,64,33,240,80,97,37,32,42,96,65,76,37,76,65,96,37,192,37,64,65,64,37,236,76,47,65,32,37,172,46,37,140,65,96,37,224,64,65,64,33,134,76,70,33,38,65,224,37,160,65,64,37,32,65,224,65,32,37,35,140,34,37,32,33,64,65,64,33,240,80,97,37,172,65,64,37,32,97,64,69,192,69,32,65,44,32,37,33,96,204,65,32,37,192,37,128,161,76,33,140,79,97,128,44,32,33,192,65,39,224,128,40,97,38,37,134,65,96,40,65,32,37,192,37,32,33,64,65,64,65,240,80,65,37,160,65,64,37,32,33,128,37,224,32,37,160,37,33,224,64,65,32,37,192,37,96,40,65,140,65,108,33,108,65,96,65,64,76,65,44,97,224,192,97,32,37,224,96,65,43,37,96,42,64,37,32,33,64,65,64,65,240,80,65,37,64,42,64,65,43,32,37,32,53,128,37,42,224,37,160,37,33,42,96,55,128,65,32,37,96,42,64,37,64,43,32,65,224,192,65,75,160,44,64,42,64,65,224,192,97,224,160,97,37,193,37,65,64,65,64,193,80,65,48,225,37,129,37,225,33,37,143,225,225,97,79,97,37,225,225,33,79,65,79,32,111,225,225,225,97,96,225,143,33,47,65,47,193,37,193,37,65,64,65,64,193,80,65,48,225,37,129,37,225,33,37,225,225,225,161,37,225,225,225,32,225,225,225,193,96,225,225,225,32,37,43,64,45,64,37,32,33,64,65,128,161,144,97,128,37,32,65,32,37,64,65,96,45,32,36,96,68,64,65,224,32,45,128,65,32,37,32,68,64,36,160,65,32,68,32,100,64,68,64,65,36,160,36,64,68,32,36,65,32,97,68,96,36,45,64,65,68,224,128,65,32,76,128,76,32,33,64,65,192,35,161,112,97,128,65,32,37,64,65,224,160,40,65,39,224,224,32,37,224,96,40,65,39,224,128,40,65,32,129,224,192,33,224,32,40,65,224,64,76,64,65,224,42,64,33,64,65,192,40,193,144,65,96,65,32,37,32,40,65,224,224,224,224,96,37,224,32,42,64,65,224,192,225,33,224,128,65,32,37,97,37,96,65,224,32,43,128,65,192,65,64,33,64,65,224,64,129,176,33,96,65,32,37,64,65,224,224,224,32,37,44,65,76,225,225,97,224,192,97,132,65,32,37,44,46,76,65,46,37,33,64,37,96,37,96,65,39,192,76,96,40,65,39,128,43,96,40,33,64,65,224,96,161,112,33,96,65,32,37,64,225,34,96,35,65,37,65,140,37,70,37,160,65,68,224,128,65,224,192,65,39,128,40,33,32,37,128,65,32,37,33,64,37,96,140,65,192,43,32,36,128,65,128,76,128,33,64,65,224,64,40,161,112,65,42,128,37,64,65,128,47,96,43,64,42,37,65,128,37,64,37,224,224,160,40,65,224,192,65,96,61,64,33,38,37,102,37,65,32,37,33,108,34,224,192,76,192,65,224,96,33,42,32,65,224,128,161,112,65,32,75,32,37,32,42,65,42,96,225,33,37,65,32,42,64,37,64,37,224,224,32,74,32,43,32,65,224,64,74,64,97,37,108,37,33,32,37,96,37,65,32,37,33,64,36,224,224,68,96,47,32,47,65,224,96,65,32,65,224,192,97,144,225,129,37,49,58,59,49,58,59,49,58,59,225,129,37,32,37,108,97,108,33,76,65,140,97,108,33,76,225,33,45,37,96,37,33,32,37,33,32,33,37,65,32,37,33,43,192,37,129,44,64,76,160,33,32,97,32,76,166,37,65,64,65,224,192,97,240,240,208,33,240,48,33,176,65,32,37,96,65,224,224,224,224,65,32,37,96,37,33,32,37,33,32,33,32,65,32,37,33,76,37,128,37,64,65,96,36,224,64,65,42,64,43,32,43,64,37,32,33,64,65,64,43,128,43,64,106,97,240,240,208,33,240,112,97,48,65,32,37,64,42,65,32,42,224,224,32,43,224,32,74,65,32,37,96,37,53,32,37,33,32,33,32,65,32,37,33,43,32,37,128,37,64,65,224,96,47,32,47,97,79,97,47,33,37,65,64,65,108,33,76,225,225,225,225,225,225,33,112,225,225,225,225,37,33,37,225,225,97,32,161,37,225,225,111,65,79,97,32,225,129,37,65,64,65,192,225,225,225,225,225,225,33,112,225,225,225,225,37,33,37,225,225,97,32,161,37,225,225,225,97,32,97,160,43,42,32,37,32,33,64,65,224,192,65,64,132,32,45,32,100,32,65,224,64,45,96,65,64,65,144,33,45,32,43,42,65,192,45,192,65,128,36,37,33,37,32,100,160,37,192,33,32,33,32,65,32,37,96,45,224,65,224,45,64,33,64,65,128,37,44,33,46,44,32,33,64,65,96,111,224,224,224,224,224,64,65,32,42,33,144,65,32,37,129,224,192,65,39,128,37,33,37,160,65,76,37,70,37,102,33,32,33,32,65,32,37,224,128,65,32,44,224,32,33,64,65,39,32,37,76,32,45,96,33,224,35,33,34,128,42,128,43,224,128,42,160,74,43,192,40,65,37,65,144,33,64,37,64,65,224,33,134,97,224,192,65,160,37,43,64,33,32,33,32,65,32,37,172,37,128,40,65,32,172,37,96,36,64,65,64,37,160,35,44,33,224,37,96,43,37,108,37,172,37,192,37,236,108,224,65,37,33,144,65,39,32,37,64,65,224,192,65,224,47,160,65,39,64,43,32,37,65,160,65,32,37,160,37,160,65,64,100,32,37,192,65,64,37,224,33,224,37,35,34,32,76,32,45,32,37,160,37,128,42,32,37,224,32,76,224,65,37,33,144,65,64,37,64,65,64,33,96,74,96,37,32,65,39,192,33,160,65,64,97,37,32,41,160,65,32,37,64,43,32,43,37,128,40,65,39,160,37,192,65,39,32,37,230,33,32,42,65,96,37,32,42,192,37,160,37,32,35,140,37,224,32,76,39,192,65,37,65,144,65,32,37,192,33,46,161,64,37,32,65,128,49,58,59,49,58,59,49,58,59,65,128,33,37,64,43,128,65,32,37,108,46,76,34,128,65,128,37,108,37,128,65,64,37,224,33,76,65,34,37,236,37,64,37,160,37,96,68,32,37,224,32,68,224,64,37,32,35,33,144,33,32,37,128,43,96,161,39,32,37,32,65,75,42,33,240,80,65,39,96,161,32,33,32,40,65,32,37,128,68,160,65,96,43,37,96,37,32,35,129,64,37,192,65,160,37,224,37,204,37,32,37,224,224,224,160,37,64,33,144,225,97,49,58,59,33,49,58,59,49,58,59,193,240,80,65,224,192,65,32,37,224,224,32,76,37,96,37,128,65,64,37,192,41,192,37,224,37,192,37,32,37,224,224,224,160,37,64,35,33,240,240,80,33,240,240,240,65,224,192,65,32,37,224,224,96,37,96,37,108,37,65,32,43,37,224,192,37,42,32,43,128,37,43,32,42,96,37,32,37,224,160,43,32,43,224,128,37,96,33,240,240,80,33,240,240,240,65,224,192,65,32,37,224,224,96,37,224,37,65,47,225,225,225,225,97,47,225,225,225,225,225,225,225,225,225,225,97,143,129,47,33,47,225,225,225,225,129,0};
    return decompMap(map, map_C, 0);
}


uint8_t* initMap()
{
    uint8_t* map = (uint8_t*)myMalloc(MAP_SIZE);
    uint8_t* ptr=map;
    memset(ptr, 1, MAP_SIZE);
    ptr=initMap1(ptr);
    ptr=initMap2(ptr);
    return map;
}

void initTiles1(uint8_t** tiles)
{
    uint8_t i;
    uint8_t tiles_C[] = {240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,16,0, // 0:rien
                         20,28,20,28,20,28,20,92,20,28,20,44,20,28,20,28,20,28,20,28,20,28,20,28,20,44,20,28,20,28,20,28,20,28,20,28,20,28,20,76,20,60,20,28,20,28,20,28,20,28,20,76,20,28,20,28,20,60,20,28,20,28,20,44,20,28,20,28,20,60,20,28,20,28,20,76,20,60,20,28,20,28,20,28,20,28,20,44,20,28,20,28,20,28,20,28,20,28,20,28,20,44,20,28,20,28,20,28,20,28,20,28,20,28,20,28,20,44,20,28,20,28,20,28,20,44,20,92,20,60,20,28,20,44,20,28,20,28,20,60,20,28,20,28,20,28,20,44,20,76,20,44,20,28,20,60,20,44,20,28,20,60,20,28,20,28,20,44,20,28,20,28,20,28,20,60,20,28,20,44,20,28,20,28,20,28,20,60,20,76,20,28,20,60,20,28,20,28,0, // 1:bloc
                         20,28,20,28,20,28,20,92,20,28,20,44,20,28,20,28,20,28,20,28,20,28,20,28,20,44,20,28,20,28,20,28,20,28,16,140,20,60,20,28,20,108,32,44,20,28,20,28,20,92,64,28,20,28,20,28,20,92,80,44,20,60,20,60,96,28,20,28,20,28,20,28,20,28,112,20,28,20,28,20,28,20,28,128,140,128,124,144,124,144,108,160,108,160,92,176,92,176,0, // 2:vouted
                         204,20,28,20,44,20,28,20,60,20,28,20,28,20,28,20,172,20,28,20,28,20,60,32,124,20,28,20,28,20,44,64,108,20,28,20,28,20,28,80,76,20,28,20,28,20,44,96,44,20,28,20,28,20,28,20,28,112,44,20,28,20,28,20,44,112,28,20,28,20,28,20,28,20,28,128,44,20,92,144,124,144,124,144,124,160,108,160,108,160,108,0, // 3:blocg
                         182,28,246,214,142,134,142,134,142,16,38,16,70,126,32,22,48,54,126,32,22,48,54,62,16,46,48,22,48,54,46,48,30,112,54,46,48,30,128,38,46,48,30,128,38,30,64,30,128,38,30,224,22,30,224,22,30,224,22,30,112,0, // 4:stala
                         80,44,48,34,144,44,48,34,128,163,80,92,88,128,44,48,34,144,44,48,34,144,44,48,34,128,163,80,92,88,128,44,48,34,144,44,48,34,144,44,48,34,128,163,80,92,88,128,44,48,34,144,44,48,34,64,0, // 5:échelle
                         240,240,240,240,240,240,240,240,128,246,22,240,240,240,240,240,240,240,112,0, // 6:corde
                         144,17,240,16,23,192,17,240,32,47,16,31,160,21,32,31,160,21,80,17,176,31,51,144,164,112,52,16,20,16,36,144,108,176,76,80,18,96,18,38,96,162,96,18,128,18,224,34,240,18,96,0, // 7:torched1
                         96,23,240,240,17,16,17,192,17,224,17,16,33,16,31,192,27,16,27,192,83,144,164,112,36,16,20,16,52,144,108,176,76,208,44,18,96,18,96,162,96,18,128,18,96,34,224,18,144,0, // 8:torcheg1
                         252,28,32,37,48,37,48,37,128,69,176,101,144,76,70,112,28,16,28,16,28,22,16,22,16,22,80,198,48,236,16,108,64,108,82,96,146,128,66,16,34,160,34,48,18,160,18,240,240,240,80,0, // 9:fontaine-anim1
                         240,240,240,240,240,240,240,240,224,71,176,22,16,22,16,39,144,22,16,22,16,30,16,46,112,22,48,22,92,112,30,16,30,16,76,144,30,16,30,54,176,46,38,208,30,22,112,0, // 10:diamant
                         240,240,96,76,62,128,44,96,30,96,28,32,60,32,30,112,28,240,16,92,61,128,44,96,30,96,28,32,60,32,30,112,28,240,16,92,61,128,44,96,30,96,28,32,60,32,30,112,28,160,30,80,44,96,46,128,60,62,80,0, // 11:ressort
                         20,22,20,22,20,22,20,86,20,22,20,38,20,22,20,22,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,20,22,20,22,20,70,20,54,20,22,20,22,20,22,20,22,20,70,20,22,20,22,20,54,20,22,20,22,20,38,20,22,20,22,20,54,20,22,20,22,20,70,20,54,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,28,54,28,22,28,20,54,20,22,20,22,28,38,28,38,44,22,20,22,20,22,20,22,20,38,28,38,60,38,20,22,20,54,140,22,20,22,20,22,20,38,140,38,20,22,20,38,20,28,22,252,124,0, // 12:mou
                         16,116,124,64,92,84,128,52,60,176,38,44,160,44,64,38,96,44,128,38,64,28,48,44,38,48,22,48,28,48,28,64,22,48,22,32,28,32,28,96,22,32,22,32,28,32,28,32,28,22,32,22,32,22,32,28,32,28,64,22,16,22,32,22,32,28,48,28,48,22,64,22,48,28,48,44,22,64,22,64,44,128,38,96,44,64,38,160,44,38,96,0, // 13:spirale1
                         20,28,20,28,20,28,20,92,20,28,20,44,20,28,20,28,20,28,20,28,20,28,20,28,20,44,20,28,20,28,20,28,20,28,20,28,20,28,20,76,20,60,20,28,20,28,20,28,20,28,20,76,20,28,20,28,20,60,20,28,20,28,20,44,20,28,20,28,20,60,20,28,20,28,20,76,20,60,20,28,20,28,20,28,20,28,20,44,20,28,20,28,20,28,20,28,20,28,20,28,20,44,20,28,20,28,20,28,20,28,20,28,20,28,20,28,20,44,20,28,20,28,20,28,20,44,20,92,20,60,20,28,20,44,20,28,20,28,20,60,20,28,20,28,20,28,20,44,20,76,20,44,20,28,20,60,20,44,20,28,20,60,20,28,20,28,20,44,20,28,20,28,20,28,20,60,20,28,20,44,20,28,20,28,20,28,20,60,20,76,20,28,20,60,20,28,20,28,0, // 14:trappe
                         240,80,17,112,17,112,17,112,17,112,17,112,17,112,17,112,17,96,49,80,17,16,17,80,49,96,33,80,17,16,17,80,17,16,17,64,17,16,49,48,17,16,49,48,33,16,33,48,81,32,113,16,113,20,108,20,108,36,92,36,220,20,44,20,92,36,60,36,60,20,76,20,140,36,60,36,76,0, // 15:pic
                         244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,20,0, //16:eau
                         240,240,240,240,240,240,240,240,240,240,240,240,192,36,16,20,32,68,16,52,32,68,16,20,16,20,16,68,16,36,22,84,22,28,36,60,100,44,68,28,100,0, // 17:surface1anim1
                         240,240,240,176,35,208,19,32,28,176,19,64,28,176,19,32,28,208,44,208,18,52,176,34,20,32,20,144,50,20,48,20,112,97,32,33,96,66,52,32,20,96,97,32,33,128,34,20,32,20,176,65,80,0, // 18:fiole
                         240,240,240,240,240,240,240,240,240,96,63,176,95,160,111,144,127,128,143,112,159,96,127,144,111,64,0, // 19:portail
                         80,127,144,111,160,111,160,111,160,111,160,111,160,111,160,111,64,31,80,111,48,31,96,111,32,31,112,111,16,31,128,111,160,111,160,52,54,112,102,100,48,116,118,16,0, // 20:portail
                         80,62,54,112,94,16,108,48,62,32,30,16,22,32,70,16,108,48,116,28,224,44,32,28,32,28,32,28,32,28,32,28,22,16,102,108,16,28,22,32,22,32,22,32,28,32,28,32,28,22,32,22,32,22,32,28,32,28,32,28,38,16,22,32,22,32,22,32,22,32,54,16,22,32,22,32,22,32,22,32,54,16,22,32,22,32,28,32,28,32,44,32,28,32,28,32,28,32,28,32,44,32,28,32,28,32,22,32,22,32,22,28,16,60,16,44,16,38,16,38,16,22,28,224,22,0, // 21:grille
                         80,104,144,88,16,40,112,87,40,16,40,96,119,16,39,96,119,16,39,96,119,16,39,96,103,16,55,112,71,16,55,144,103,160,51,63,176,35,63,160,83,192,35,63,160,51,47,192,35,63,192,47,112,0, // 22:lampe
                         80,44,16,60,128,76,16,92,80,76,16,124,48,60,32,44,119,32,39,16,183,16,39,16,55,64,103,32,55,96,167,96,167,96,167,96,183,64,39,48,23,16,167,16,55,32,119,48,71,48,87,16,108,80,60,16,108,128,28,16,76,80,0, // 23:bouée
                         240,240,240,240,240,240,240,240,128,63,208,95,176,111,160,127,144,143,128,159,160,127,160,111,96,0, // 24:portail
                         80,111,160,111,160,111,160,111,160,111,160,111,160,111,80,31,64,111,96,31,48,111,112,31,32,111,128,31,16,111,160,111,160,111,160,52,54,112,102,100,48,116,118,16,0, // 25:portail
                         240,240,240,240,240,240,240,240,240,240,240,240,192,36,64,68,16,52,32,68,16,20,16,20,16,68,16,36,22,84,22,28,36,60,100,44,68,28,100,0, // 27:surface2anim1
                         240,240,240,240,240,240,240,240,240,240,240,240,240,32,36,112,20,32,20,16,52,32,52,16,36,32,20,22,36,22,68,30,52,30,20,16,36,22,20,22,52,30,100,16,0, // 28:surface3anim1
                         240,240,128,85,160,51,16,53,144,19,80,17,144,19,80,17,144,51,16,49,160,35,57,192,19,25,224,19,25,224,19,70,176,19,22,224,19,54,192,19,22,224,19,70,240,80,0, // 28:clef
                         80,19,16,35,16,19,160,19,16,35,16,19,176,73,176,41,16,25,16,25,176,73,176,28,16,44,16,28,144,54,32,60,112,54,16,22,28,16,60,96,54,16,22,28,16,60,96,22,32,38,92,112,38,32,76,144,22,64,28,160,38,32,44,176,22,16,44,176,54,16,34,144,50,32,50,64,0, // 29:statuette
                        };
    uint8_t* ptr_in = tiles_C;
    for(i=0; i<30; i++)
    {
       tiles[i] = myMalloc(TILE_SIZE*TILE_SIZE);
       ptr_in = decompTile(tiles[i], ptr_in);
    }
}

void initTiles2(uint8_t** tiles)
{
    uint8_t i;
    uint8_t tiles_C[] = {240,240,240,64,187,64,24,32,40,16,40,91,48,24,16,24,160,27,32,40,16,24,48,120,48,43,16,27,16,27,112,24,80,43,16,27,120,80,43,128,24,96,40,16,88,16,40,80,40,48,33,160,40,16,27,32,43,16,27,64,43,128,43,64,59,16,122,112,43,128,0, // 30:parchemin
                        240,240,48,19,64,19,32,99,32,19,64,19,32,99,32,35,32,35,32,19,112,35,32,35,32,19,112,19,16,35,16,19,32,67,64,19,16,35,16,19,32,67,64,19,64,19,32,19,112,19,64,19,32,19,112,19,64,19,32,19,112,19,64,19,32,19,112,19,64,19,32,99,32,19,64,19,32,99,240,240,48,0,
                        240,240,48,19,128,51,64,19,128,51,64,19,144,19,80,19,144,19,80,19,144,19,80,19,144,19,80,19,144,19,80,19,144,19,80,19,144,19,80,19,144,19,80,99,48,51,64,99,48,51,240,240,80,0,
                        240,240,64,67,48,19,64,19,48,67,48,19,64,19,32,19,112,19,64,19,32,19,112,19,64,19,48,67,48,99,48,67,48,99,112,19,32,19,64,19,112,19,32,19,64,19,32,19,64,19,32,19,64,19,32,19,64,19,32,19,64,19,48,67,48,19,64,19,48,67,48,19,64,19,240,240,48,0,
                        240,240,64,51,144,19,48,51,144,19,64,19,160,19,64,19,160,19,64,19,160,19,64,19,160,19,64,19,160,19,64,19,160,19,64,19,80,19,64,19,64,19,80,19,64,19,48,51,80,67,64,51,80,67,240,240,64,0,
                        240,240,80,35,64,19,48,19,80,35,64,19,48,19,64,19,32,19,48,19,32,19,80,19,32,19,48,19,32,19,64,19,64,19,32,51,80,19,64,19,32,51,80,99,32,19,16,19,80,99,32,19,16,19,80,19,64,19,32,19,32,19,64,19,64,19,32,19,32,19,64,19,64,19,32,19,48,19,48,19,64,19,32,19,48,19,240,240,64,0,                         16,116,124,64,92,84,128,52,60,176,38,44,160,38,64,44,96,38,128,44,64,22,48,38,44,48,28,48,22,48,22,64,28,48,28,32,22,32,22,96,28,32,28,32,22,32,22,32,22,28,32,28,32,28,32,22,32,22,64,28,16,28,32,28,32,22,48,22,48,28,64,28,48,22,48,38,28,64,28,64,38,128,44,96,38,64,44,160,38,44,96,0, // 36:spirale2
                         144,31,208,17,240,32,22,240,192,17,16,17,32,19,160,21,48,19,16,19,176,79,144,164,112,52,16,20,16,36,144,108,176,76,80,18,96,18,38,96,162,96,18,128,18,224,34,240,18,96,0, // 37:torched2
                         144,17,240,240,19,240,16,17,16,17,144,17,16,17,32,22,160,19,32,19,224,31,49,144,164,112,52,16,20,16,36,144,108,176,76,80,18,96,18,38,96,162,96,18,128,18,224,34,240,18,96,0, // 38:torched3
                         80,17,240,16,19,224,17,16,17,192,31,32,31,17,176,17,16,33,208,35,208,83,144,164,112,36,16,20,16,52,144,108,176,76,208,44,18,96,18,96,162,96,18,128,18,96,34,224,18,144,0, // torcheg2
                         96,17,224,23,240,32,17,208,31,240,16,31,16,23,192,23,224,83,144,164,112,36,16,20,16,52,144,108,176,76,208,44,18,96,18,96,162,96,18,128,18,96,34,224,18,144,0, // torcheg3
                         252,28,32,37,48,37,48,37,128,69,176,100,144,132,112,28,16,28,16,28,22,16,22,16,22,80,198,48,236,16,108,64,108,82,96,146,128,66,16,34,160,34,48,18,160,18,240,240,240,80,0, // 41:fontaine2
                         252,28,32,37,48,37,48,37,128,69,176,101,144,76,70,112,28,16,28,16,28,22,16,22,16,22,80,196,48,228,16,108,64,108,82,96,146,128,66,16,34,160,34,48,18,160,18,240,240,240,80,0, // 42:fontaine3
                         252,28,32,37,48,37,48,37,128,69,176,101,144,76,70,112,28,16,28,16,28,22,16,22,16,22,80,198,48,236,16,108,64,108,84,96,148,128,68,16,34,160,34,48,18,160,18,240,240,240,80,0, // 43:fontaine4
                         240,240,240,240,240,240,240,240,240,240,240,240,192,20,16,36,64,36,32,20,16,20,32,180,16,20,16,52,22,36,22,36,28,36,28,116,28,100,28,68,28,0, // 44:surface1anim2
                         240,240,240,240,240,240,240,240,240,240,240,240,224,20,16,20,48,36,48,52,16,52,16,36,16,20,16,68,16,196,28,84,28,20,28,36,28,100,28,36,0, // 45:surface1anim3
                         240,240,240,240,240,240,240,240,240,240,240,240,240,48,20,144,36,16,36,16,20,16,20,16,36,32,196,28,148,28,132,28,36,0, // 46:surface2anim2
                         240,240,240,240,240,240,240,240,240,240,240,240,224,36,80,36,48,20,16,84,16,20,16,20,16,132,22,100,28,36,28,36,28,20,28,52,28,36,28,84,28,20,0, // 47:surface2anim3
                         240,240,240,240,240,240,240,240,240,240,240,240,208,20,64,52,16,20,64,68,16,20,16,100,16,68,22,116,30,52,30,116,22,148,30,0, // 48:surface3anim2
                         240,240,240,240,240,240,240,240,240,240,240,240,240,32,20,16,20,32,20,64,36,16,20,16,68,16,52,16,36,16,84,22,68,30,116,22,68,22,52,30,52,30,0, // 49:surface3anim3
                         80,22,48,22,208,22,208,22,80,22,64,22,128,22,80,54,16,22,176,38,16,22,16,22,48,22,16,22,48,70,20,54,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,28,54,28,22,28,20,54,20,22,20,22,28,38,28,38,44,22,20,22,20,22,20,22,20,38,28,38,60,38,20,22,20,54,140,22,20,22,20,22,20,38,140,38,20,22,20,38,20,28,22,252,124,0, // 50 mou-vidage-anim1
                         240,240,176,22,128,22,80,22,32,22,144,22,112,22,32,22,96,22,32,22,64,54,208,38,16,22,16,22,64,22,64,38,20,22,20,22,20,22,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,28,54,28,22,28,20,54,20,22,20,22,28,38,28,38,44,22,20,22,20,22,20,22,20,38,28,38,60,38,20,22,20,54,140,22,20,22,20,22,20,38,140,38,20,22,20,38,20,28,22,252,124,0,// 51 mou-vidage-anim2
                         240,240,240,192,22,208,22,176,22,96,22,80,22,224,38,64,22,32,22,80,38,16,22,128,22,32,22,16,70,16,22,112,22,16,22,16,70,16,22,16,22,64,86,20,22,20,22,20,22,20,38,28,38,60,38,20,22,20,54,140,22,20,22,20,22,20,38,140,38,20,22,20,38,20,28,22,252,124,0,// 52 mou-vidage-anim3
                         240,240,240,240,240,176,22,96,22,176,22,192,22,208,22,112,22,48,22,96,22,128,38,128,22,64,70,176,102,16,22,80,28,16,60,86,16,38,140,38,20,22,20,38,20,28,22,252,124,0,// 53 mou-vidage-anim4
                         240,240,240,240,240,240,240,240,240,240,28,80,28,96,28,64,28,192,28,64,28,80,22,112,28,80,44,38,160,76,38,20,22,20,38,20,28,22,252,124,0,// 54 mou-vidage-anim5
                         240,240,240,240,240,240,240,240,240,240,28,80,28,96,28,64,28,192,28,64,28,80,22,112,28,80,44,38,160,76,38,20,22,20,38,20,28,22,252,124,0,// 55 mou-vidage-anim5
                         240,240,240,240,240,176,22,96,22,176,22,192,22,208,22,112,22,48,22,96,22,128,38,128,22,64,70,176,102,16,22,80,28,16,60,86,16,38,140,38,20,22,20,38,20,28,22,252,124,0,// 56 mou-vidage-anim4
                         240,240,240,192,22,208,22,176,22,96,22,80,22,224,38,64,22,32,22,80,38,16,22,128,22,32,22,16,70,16,22,112,22,16,22,16,70,16,22,16,22,64,86,20,22,20,22,20,22,20,38,28,38,60,38,20,22,20,54,140,22,20,22,20,22,20,38,140,38,20,22,20,38,20,28,22,252,124,0,// 57 mou-vidage-anim3
                         240,240,176,22,128,22,80,22,32,22,144,22,112,22,32,22,96,22,32,22,64,54,208,38,16,22,16,22,64,22,64,38,20,22,20,22,20,22,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,28,54,28,22,28,20,54,20,22,20,22,28,38,28,38,44,22,20,22,20,22,20,22,20,38,28,38,60,38,20,22,20,54,140,22,20,22,20,22,20,38,140,38,20,22,20,38,20,28,22,252,124,0,// 58 mou-vidage-anim2
                         80,22,48,22,208,22,208,22,80,22,64,22,128,22,80,54,16,22,176,38,16,22,16,22,48,22,16,22,48,70,20,54,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,20,22,20,22,20,22,20,38,20,22,20,22,20,22,20,22,28,54,28,22,28,20,54,20,22,20,22,28,38,28,38,44,22,20,22,20,22,20,22,20,38,28,38,60,38,20,22,20,54,140,22,20,22,20,22,20,38,140,38,20,22,20,38,20,28,22,252,124,0, // 59 mou-vidage-anim1
                        };
    uint8_t* ptr_in = tiles_C;
    for(i=30; i<NTILES-1; i++)
    {
       tiles[i] = myMalloc(TILE_SIZE*TILE_SIZE);
       ptr_in = decompTile(tiles[i], ptr_in);
    }
    tiles[i] = myMalloc(TILE_SIZE*TILE_SIZE);
    memset(tiles[i],16,TILE_SIZE*TILE_SIZE);
}

void initTilesP(uint8_t** tilesp)
{
    uint8_t i;
    uint8_t tilesp_C[] = {
        240,112,18,16,18,32,18,112,34,48,34,48,18,112,50,63,16,31,112,143,112,47,32,47,32,47,96,47,32,47,32,47,96,175,96,47,16,79,16,47,112,47,64,47,96,31,32,111,32,31,80,28,32,76,32,28,112,28,32,44,32,28,144,102,160,102,160,33,32,33,80,0, // 0:defaut-anim1
        80,18,32,18,16,18,16,18,112,18,16,18,32,18,16,18,144,50,63,16,31,112,143,112,47,32,47,32,47,96,47,32,47,32,47,96,31,16,111,16,31,96,47,16,79,16,47,112,47,64,47,144,111,176,76,160,28,32,44,32,28,112,22,16,102,16,22,128,102,160,38,32,38,144,33,64,33,64,0, // 1:defaut-anim2
        96,18,48,50,112,34,16,18,16,18,16,18,16,18,112,63,66,128,143,112,63,32,95,80,79,32,95,80,191,96,175,112,143,144,111,192,76,208,44,160,22,16,86,176,102,144,38,48,38,16,22,112,49,32,49,48,0, // 2:gauche-anim1
        240,96,34,32,34,144,18,32,18,16,18,16,34,128,63,50,144,143,112,63,32,95,80,79,32,95,80,191,96,175,112,143,144,111,176,76,160,28,32,44,192,86,176,102,144,49,16,49,80,0, // 3:gauche-anim2
        96,18,48,18,144,18,32,18,16,18,16,18,112,18,16,50,79,144,143,112,111,32,47,96,111,32,63,80,191,80,175,112,143,144,111,160,76,208,44,208,86,16,22,128,102,128,22,16,38,48,38,128,49,16,33,80,0, // 4:droite-anim1
        240,64,18,32,34,32,34,128,18,16,18,32,18,144,18,16,50,63,144,143,112,111,32,47,96,111,32,63,80,191,80,175,112,143,144,111,176,76,208,44,32,28,160,86,160,54,16,38,176,49,16,49,64,0, // 5:droite-anim2
        80,34,32,34,144,18,32,18,16,18,16,34,112,79,50,128,159,96,191,80,191,80,191,80,191,96,159,128,127,32,31,112,92,32,28,96,44,32,44,32,28,144,102,160,134,112,38,64,33,128,33,160,0, // 6:échelle-anim1
        48,18,32,34,32,18,16,18,112,18,32,18,16,18,16,18,144,63,66,128,159,96,191,80,191,80,191,80,191,96,159,80,31,32,127,112,28,32,92,144,28,32,44,32,44,128,102,128,150,128,33,64,44,224,33,64,0, // 7:échelle-anim2
        48,18,32,18,16,18,16,18,144,18,32,18,16,18,16,34,128,111,144,143,112,175,96,175,96,175,96,175,48,22,31,38,143,38,31,22,16,47,32,111,32,47,48,44,32,76,32,28,96,44,32,44,16,44,144,102,160,102,160,38,32,38,160,33,32,33,80,0, // 8:corde
        48,18,16,18,32,18,16,18,144,18,32,18,16,18,16,34,128,63,50,144,143,112,63,32,95,80,79,32,95,80,191,96,175,48,22,31,38,143,70,16,47,32,111,112,44,32,76,32,60,64,44,32,44,16,44,144,86,176,102,160,38,32,38,144,49,16,49,80,0, // 9:corde-gauche-anim1
        240,48,18,48,18,48,18,16,18,96,34,16,34,16,18,16,18,144,63,50,144,143,112,63,32,95,80,79,32,95,80,191,48,54,175,54,64,143,32,31,96,111,32,47,32,60,32,76,32,44,80,44,32,44,16,60,112,102,160,102,176,33,32,33,80,0, // 10:corde-gauche-anim2
        240,96,18,16,18,160,18,48,18,32,18,16,18,112,18,16,50,63,144,143,112,111,32,47,96,111,32,63,80,191,32,54,175,38,31,64,143,48,31,80,111,48,47,16,44,48,76,48,44,48,44,48,44,32,60,96,22,16,86,176,86,160,33,32,49,64,0, // 11:corde-droite-anim1
        80,18,16,18,32,18,112,18,16,18,16,18,32,18,16,18,112,18,16,50,63,144,143,112,111,32,47,96,111,32,63,80,191,48,31,16,175,48,38,31,22,143,70,32,31,32,111,112,44,32,76,80,28,48,44,32,44,32,28,32,44,64,22,16,86,16,54,96,102,160,38,32,38,144,49,16,49,80,0, // 12:corde-droite-anim2
        84,18,36,18,52,18,116,18,20,18,36,18,20,18,148,50,63,20,31,116,143,116,175,100,47,36,47,36,47,100,175,100,175,116,47,68,47,100,31,36,111,36,31,84,28,36,76,36,28,116,28,36,44,36,28,148,102,164,102,116,17,36,33,36,33,36,17,84,49,68,49,52,0, // 13:nage-anim1
        68,18,20,18,36,18,20,18,148,50,63,20,31,116,143,116,175,100,47,36,47,36,47,100,175,100,175,116,63,36,63,148,111,180,76,164,28,36,44,36,28,116,22,20,102,20,22,132,102,164,33,36,33,116,81,36,81,244,52,0, // 14:nage-anim2
        48,18,16,18,32,18,16,18,16,18,64,18,16,18,32,18,32,18,16,18,16,18,64,18,16,18,16,18,32,34,16,18,96,47,32,47,48,31,96,175,96,47,16,63,16,63,96,47,32,47,32,47,96,47,32,47,32,47,112,143,96,31,32,47,32,47,32,31,80,28,32,76,32,28,112,28,32,44,32,28,144,102,160,102,144,54,32,54,144,17,64,17,80,0, // 15:tombe
        240,96,111,144,143,112,31,16,31,16,47,16,31,16,31,96,47,16,79,16,47,96,175,48,31,32,63,16,47,16,63,32,47,48,63,32,63,48,63,48,111,48,47,76,32,76,32,76,16,76,32,44,32,76,48,38,16,102,16,38,112,102,160,102,160,38,32,38,144,49,32,49,240,80,0, // 16:ange-anim1
        240,240,112,111,144,143,112,31,16,31,16,47,16,31,16,31,96,47,16,79,16,47,96,175,96,175,112,63,32,63,144,111,112,44,32,76,32,44,48,76,32,44,32,76,32,54,16,102,16,54,16,38,48,102,48,54,64,38,32,38,64,22,17,64,33,32,33,80,0, // 17:ange-anim2
        48,143,112,161,80,193,48,65,16,145,32,49,16,33,16,129,16,241,97,16,241,33,16,177,16,17,16,241,177,16,33,16,65,16,28,16,76,48,60,32,60,16,28,16,28,64,28,16,28,32,28,16,28,32,28,16,28,48,28,16,28,32,28,16,28,32,28,32,28,32,28,16,28,16,28,48,28,16,28,32,28,16,28,32,28,16,28,48,28,16,28,32,28,0, // 18:medusoide-anim1
        80,95,160,113,128,145,96,33,16,129,64,81,16,113,32,193,16,33,16,49,16,17,16,17,16,113,16,241,65,16,241,209,16,33,16,65,16,17,16,65,16,28,16,44,16,28,16,60,16,28,16,28,32,28,16,28,32,28,16,28,16,28,32,28,16,28,48,28,32,28,16,28,16,28,32,28,16,28,48,28,32,28,16,28,32,28,16,28,16,28,48,28,80,44,32,28,32,28,0, // 19:medusoide-anim2
        240,96,111,128,145,96,65,16,97,64,209,32,129,16,97,16,81,16,241,33,16,177,16,17,16,241,161,32,33,16,65,16,17,16,65,48,60,32,28,33,16,17,16,17,64,28,16,28,32,28,16,28,32,28,16,28,48,28,16,28,32,28,16,28,32,28,16,28,48,28,64,28,32,28,16,28,32,28,48,28,80,28,16,28,32,28,16,0, // 20:medusoide-anim3
    };
    uint8_t* ptr_in = tilesp_C;
    for(i=0; i<NTILESP; i++)
    {
       tilesp[i] = myMalloc(TILE_SIZE*TILE_SIZE);
       ptr_in = decompTile(tilesp[i], ptr_in);
    }
}

/* reverse string s in place */
void reverse(char* s)
{
    uint16_t i, j;
    char c;

    for (i = 0, j = strlen(s)-1; i<j; i++, j--) {
        c = s[i];
        s[i] = s[j];
        s[j] = c;
    }
}

// Unsigned int to string (s), which also returns the pointer to the string
char* uitoa(uint24_t n, char* s, uint8_t zeropad_len)
{
    uint24_t i = 0;

    do {       /* generate digits in reverse order */
        s[i++] = n % 10 + '0';   /* get next digit */
    } while ((n /= 10) > 0);     /* delete it */

    for(; i<zeropad_len; i++)
        s[i] = '0';

    s[i] = '\0';

    reverse(s);

    return s;
}

void* myMalloc(size_t size)
{
    if (safeRAM_free > size)
    {
        safeRAM_ptr_old = safeRAM_ptr;
        safeRAM_ptr += size;
        safeRAM_free -= size;
        return safeRAM_ptr-size;
    } else {
        cleanUp();
        printSmall8bpp("No RAM left... :(", 0, 0);
        asm("jp _RestoreStackAndExit"); // Force quit program
        return 0;
    }
}

void myFreeLast() {
    safeRAM_free += safeRAM_ptr - safeRAM_ptr_old;
    safeRAM_ptr = safeRAM_ptr_old;
}

void initPalette()
{
    palette[0] = rgb2color(0, 0, 0);        // noir
    palette[1] = rgb2color(255, 0, 0);    // rouge
    palette[2] = rgb2color(0, 255, 0);      // vert
    palette[3] = rgb2color(255, 255, 0);    // jaune
    palette[4] = rgb2color(0, 0, 255);    // bleu
    palette[5] = rgb2color(255, 0, 255);    // rose
    palette[6] = rgb2color(0, 255, 255);    // cyan
    palette[7] = rgb2color(255, 255, 255);  // blanc
    palette[8] = rgb2color(160, 160, 160);  // gris
    palette[9] = rgb2color(255, 160, 160);  // rouge clair
    palette[10] = rgb2color(160, 255, 160); // vert clair
    palette[11] = rgb2color(255, 255, 160); // jaune clair
    palette[12] = rgb2color(64, 64, 255);  // bleu clair
    palette[13] = rgb2color(255, 160, 255); // rose clair
    palette[14] = rgb2color(96, 255, 255); // cyan clair
    palette[15] = rgb2color(255, 160, 40);  // orange
    palette[16] = rgb2color(32, 192, 224);
    palette[17] = rgb2color(128, 128, 128);
}

void setAutoCanvas()
{
    uint8_t x0=pers[0].x - SCR_COLS/2;
    uint8_t y0=pers[0].y - SCR_LINES/2, t=0;
    if(pers[0].force && pers[0].y<pers[0].yold && !canHang(0,pers[0].idtile) && !pers[0].jumps)
    {
        t=(pers[0].y<MAP_HEIGHT-1)?getTile(*(getMapTilePtr(pers[0].x, pers[0].y+1))):1;
        if(canGo(0,t,pers[0].x, pers[0].y+1) && t!=5)
            y0++;
    }
    if(x>x0 || x0>=MAP_WIDTH) x--;
    else if(x<x0) x++;
    if(y>y0 || y0>=MAP_HEIGHT) y--;
    else if(y<y0) y++;
    if (x>=MAP_WIDTH-SCR_COLS+1)
        x = xold;
    if (y>=MAP_HEIGHT-SCR_LINES+1)
        y = yold;
}

uint8_t* decompTile(uint8_t* data_out, uint8_t* data_in)
{
    uint8_t n=0, colid=0;
    while (*data_in>0)
    {
        n = *data_in >> 4;
        colid = *data_in%16;
        memset(data_out, colid, n);
        data_out += n;
        data_in++;
    }
    return data_in+1;
}

void gc_drawTile(uint8_t* data, uint16_t x, uint16_t y, uint8_t w, uint8_t h)
{
    uint8_t* ptr_out = screen_buf;
    uint8_t* ptr_in  = data;
    ptr_out += y*SCR_WIDTH + x;
    while (h>0)
    {
        memcpy(ptr_out, ptr_in, w);
        ptr_out += SCR_WIDTH;
        ptr_in  += w;
        h--;
    }
}

void gc_drawTileTrsp(uint8_t* data, uint16_t x, uint16_t y, uint8_t w, uint8_t h)
{
    uint8_t* ptr_out = screen_buf;
    uint8_t* buffer = myMalloc(w*h);
    uint8_t* ptr_in  = buffer;
    uint8_t i=0,j=0;
    memcpy(buffer,data,w*h);
    ptr_out += y*SCR_WIDTH + x;
    for(j=0;j<h;j++)
        for(i=0;i<w;i++)
            if(!buffer[w*j+i])
                buffer[w*j+i]=17;
    while (h>0)
    {
        memcpy(ptr_out, ptr_in, w);
        ptr_out += SCR_WIDTH;
        ptr_in  += w;
        h--;
    }
    myFreeLast();
}

void cleanUp()
{
    asm("ld hl,0F50000h");
    asm("xor a"); // keypad idle mode
    asm("ld (hl),a");
    // Back to 320x240 16bpp
    asm("LD A,2Dh");
    asm("LD (E30018h),A");

    asm("CALL _DelRes");
    asm("CALL _ClrTxtShd");
    asm("CALL _ClrScrn");
    asm("SET 0,(iy+3)"); // mark graph as dirty
    asm("CALL _HomeUp");
    asm("CALL _DrawStatusBar");
    asm("CALL _RunIndicOn");
    asm("ei");
}

// No need to call, it's just to have the code get assembled.
// Pieces of code from MateoConLechuga's things here and there
void textRoutines()
{
    // Some required equates
    asm("posX       equ D052C6h");
    asm("posY       equ D052C9h");
    asm("_cphlde    equ 02013Ch");
    asm("vbuf1      equ D40000h");
    asm("vbuf2      equ D52C00h");

    asm("drawStringSmall:");
    asm(" ld a,(hl)");
    asm(" or a");
    asm(" ret z");
    asm(" call DrawChar");
    asm(" push hl");
    asm("  ld de,320-10");
    asm("  ld hl,(posX)");
    asm("  call _cphlde");
    asm(" pop hl");
    asm(" ret nc");
    asm(" inc hl");
    asm(" jr drawStringSmall");

    asm("drawStringLarge:");
    asm(" ld a,(hl)");
    asm(" or a");
    asm(" ret z");
    asm(" call DrawBigChar");
    asm(" push hl");
    asm("  ld de,320-10");
    asm("  ld hl,(posX)");
    asm("  call _cphlde");
    asm(" pop hl");
    asm(" ret nc");
    asm(" inc hl");
    asm(" jr drawStringLarge");

    asm("DrawBigChar:");
    asm(" ld bc,(posX)");
    asm(" push hl");
    asm(" push af");
    asm(" push de");
    asm(" push bc");
    asm(" push af");
    asm("  ld a,(posY)");
    asm("  push bc   ; Save X");
    asm("   call compute8bpp");
    asm("  pop de    ; de = Y");
    asm("  add hl,de ; Add X");
    asm(" pop af");
    asm(" push hl");
    asm("  or a");
    asm("  sbc hl,hl");
    asm("  ld l,a");
    asm("  add hl,hl");
    asm("  add hl,hl");
    asm("  add hl,hl");
    asm("  ex de,hl");
    asm("  ld hl,Char000");
    asm("  add hl,de ; hl -> Correct Character");
    asm(" pop de     ; de -> correct place to draw");
    asm(" ld b,8");
    asm("_iloop2:");
    asm(" push bc");
    asm("  ld c,(hl)");
    asm("  ld b,8");
    asm("  ex de,hl");
    asm("  push de");
    asm("  push hl");
    asm("   ld de,(ForeColor)");
    asm("_i2loop2:");
    asm("   ld a,d");
    asm("   rlc c");
    asm("   jr nc,+_heh1");
    asm("   ld a,e");
    asm("_heh1:");
    asm("   ld (hl),a");
    asm("   inc hl");
    asm("   ld (hl),a");
    asm("   inc hl");
    asm("   djnz _i2loop2");
    asm("   ld bc,320-16");
    asm("   add hl,bc    ; next line");
    asm("   ex de,hl");
    asm("   pop hl");
    asm("   ld bc,16");
    asm("   ldir");
    asm("   ex de,hl");
    asm("   ld bc,320-16");
    asm("   add hl,bc    ; next line");
    asm("  pop de");
    asm("  ex de,hl");
    asm("  inc hl");
    asm(" pop bc");
    asm(" djnz _iloop2");
    asm(" pop bc");
    asm(" pop de");
    asm(" pop af     ; character");
    asm(" cp 128");
    asm(" jr c,+_heh2");
    asm(" xor a");
    asm("_heh2:");
    asm(" ld hl,CharSpacing");
    asm(" call _AddHLAndA");
    asm(" ld a,(hl)  ; A holds the amount to increment per character");
    asm(" add a,a");
    asm(" jp AddToPosistion");

    asm("DrawChar:");
    asm(" ld bc,(posX)");
    asm(" push hl");
    asm(" push af");
    asm(" push de");
    asm(" push bc");
    asm(" push af");
    asm("  ld a,(posY)");
    asm("  push bc   ; Save X");
    asm("   call compute8bpp");
    asm("  pop de    ; de = Y");
    asm("  add hl,de ; Add X");
    asm(" pop af");
    asm(" push hl");
    asm("  or a");
    asm("  sbc hl,hl");
    asm("  ld l,a");
    asm("  add hl,hl");
    asm("  add hl,hl");
    asm("  add hl,hl");
    asm("  ex de,hl");
    asm("  ld hl,Char000");
    asm("  add hl,de ; hl -> Correct Character");
    asm(" pop de     ; de -> correct place to draw");
    asm(" ld b,8");
    asm("_iloop:");
    asm(" push bc");
    asm("  ld c,(hl)");
    asm("  ld b,8");
    asm("  ex de,hl");
    asm("  push de");
    asm("   ld de,(ForeColor)");
    asm("_i2loop:");
    asm("   ld (hl),d");
    asm("   rlc c");
    asm("   jr nc,+_heh3");
    asm("   ld (hl),e");
    asm("_heh3:");
    asm("   inc hl");
    asm("   djnz _i2loop");
    asm("   ld (hl),d");
    asm("NextL:");
    asm("   ld bc,320-8");
    asm("   add hl,bc");
    asm("  pop de");
    asm("  ex de,hl");
    asm("  inc hl");
    asm(" pop bc");
    asm(" djnz _iloop");
    asm(" pop bc");
    asm(" pop de");
    asm(" pop af     ; character");
    asm(" cp 128");
    asm(" jr c,+_");
    asm(" xor a");
    asm("_:");
    asm(" ld hl,CharSpacing");
    asm(" call _AddHLAndA");
    asm(" ld a,(hl)  ; A holds the amount to increment per character");

    asm("AddToPosistion:");
    asm(" or a,a");
    asm(" sbc hl,hl");
    asm(" ld l,a");
    asm(" add hl,bc");
    asm(" push hl");
    asm(" pop bc");
    asm(" inc bc");
    asm(" ld (posX),bc");
    asm(" pop hl");
    asm(" ret");

    asm("ForeColor:");
    asm(" DB 7");
    asm("BackColor:");
    asm(" DB 17");

    asm("compute8bpp:");
    asm(" ld de,320");
    asm(" call 000348h               ; MultDEA");
    //asm(" ld de,vbuf2");  // off screen (vram + 320x240)
    asm(" ld de,vbuf1");    // direct screen (vram)
    asm(" add hl,de");
    asm(" ret");

    asm("CharSpacing:");
    asm(";   0,1,2,3,4,5,6,7,8,9,A,B,C,D,E,F");
    asm(" DB 8,8,8,7,7,7,8,8,8,8,8,8,8,1,8,8");
    asm(" DB 7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8");
    asm(" DB 2,3,5,7,7,7,7,4,4,4,8,6,3,6,2,7");
    asm(" DB 7,6,7,7,7,7,7,7,7,7,2,3,5,6,5,6");
    asm(" DB 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7");
    asm(" DB 7,7,7,7,8,7,7,7,7,7,7,4,7,4,7,8");
    asm(" DB 3,7,7,7,7,7,7,7,7,4,7,7,4,7,7,7");
    asm(" DB 7,7,7,7,6,7,7,7,7,7,7,6,2,6,7,7");
    asm(" DB 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7");


    asm("Char000: DB 0,0,0,0,0,0,0,0");
    asm("Char001: DB 7Eh,81h,A5h,81h,BDh,BDh,81h,7Eh");
    asm("Char002: DB 7Eh,FFh,DBh,FFh,C3h,C3h,FFh,7Eh");
    asm("Char003: DB 6Ch,FEh,FEh,FEh,7Ch,38h,10h,00h");
    asm("Char004: DB 10h,38h,7Ch,FEh,7Ch,38h,10h,00h");
    asm("Char005: DB 38h,7Ch,38h,FEh,FEh,10h,10h,7Ch");
    asm("Char006: DB 00h,18h,3Ch,7Eh,FFh,7Eh,18h,7Eh");
    asm("Char007: DB 00h,00h,18h,3Ch,3Ch,18h,00h,00h");
    asm("Char008: DB FFh,FFh,E7h,C3h,C3h,E7h,FFh,FFh");
    asm("Char009: DB 00h,3Ch,66h,42h,42h,66h,3Ch,00h");
    asm("Char010: DB FFh,C3h,99h,BDh,BDh,99h,C3h,FFh");
    asm("Char011: DB 0Fh,07h,0Fh,7Dh,CCh,CCh,CCh,78h");
    asm("Char012: DB 3Ch,66h,66h,66h,3Ch,18h,7Eh,18h");
    asm("Char013: DB 3Fh,33h,3Fh,30h,30h,70h,F0h,E0h");
    asm("Char014: DB 7Fh,63h,7Fh,63h,63h,67h,E6h,C0h");
    asm("Char015: DB 99h,5Ah,3Ch,E7h,E7h,3Ch,5Ah,99h");
    asm("Char016: DB 80h,E0h,F8h,FEh,F8h,E0h,80h,00h");
    asm("Char017: DB 02h,0Eh,3Eh,FEh,3Eh,0Eh,02h,00h");
    asm("Char018: DB 18h,3Ch,7Eh,18h,18h,7Eh,3Ch,18h");
    asm("Char019: DB 66h,66h,66h,66h,66h,00h,66h,00h");
    asm("Char020: DB 7Fh,DBh,DBh,7Bh,1Bh,1Bh,1Bh,00h");
    asm("Char021: DB 3Fh,60h,7Ch,66h,66h,3Eh,06h,FCh");
    asm("Char022: DB 00h,00h,00h,00h,7Eh,7Eh,7Eh,00h");
    asm("Char023: DB 18h,3Ch,7Eh,18h,7Eh,3Ch,18h,FFh");
    asm("Char024: DB 18h,3Ch,7Eh,18h,18h,18h,18h,00h");
    asm("Char025: DB 18h,18h,18h,18h,7Eh,3Ch,18h,00h");
    asm("Char026: DB 00h,18h,0Ch,FEh,0Ch,18h,00h,00h");
    asm("Char027: DB 00h,30h,60h,FEh,60h,30h,00h,00h");
    asm("Char028: DB 00h,00h,C0h,C0h,C0h,FEh,00h,00h");
    asm("Char029: DB 00h,24h,66h,FFh,66h,24h,00h,00h");
    asm("Char030: DB 00h,18h,3Ch,7Eh,FFh,FFh,00h,00h");
    asm("Char031: DB 00h,FFh,FFh,7Eh,3Ch,18h,00h,00h");
    asm("Char032: DB 00h,00h,00h,00h,00h,00h,00h,00h");
    asm("Char033: DB C0h,C0h,C0h,C0h,C0h,00h,C0h,00h");
    asm("Char034: DB D8h,D8h,D8h,00h,00h,00h,00h,00h");
    asm("Char035: DB 6Ch,6Ch,FEh,6Ch,FEh,6Ch,6Ch,00h");
    asm("Char036: DB 18h,7Eh,C0h,7Ch,06h,FCh,18h,00h");
    asm("Char037: DB 00h,C6h,CCh,18h,30h,66h,C6h,00h");
    asm("Char038: DB 38h,6Ch,38h,76h,DCh,CCh,76h,00h");
    asm("Char039: DB 30h,30h,60h,00h,00h,00h,00h,00h");
    asm("Char040: DB 30h,60h,C0h,C0h,C0h,60h,30h,00h");
    asm("Char041: DB C0h,60h,30h,30h,30h,60h,C0h,00h");
    asm("Char042: DB 00h,66h,3Ch,FFh,3Ch,66h,00h,00h");
    asm("Char043: DB 00h,30h,30h,FCh,FCh,30h,30h,00h");
    asm("Char044: DB 00h,00h,00h,00h,00h,60h,60h,C0h");
    asm("Char045: DB 00h,00h,00h,FCh,00h,00h,00h,00h");
    asm("Char046: DB 00h,00h,00h,00h,00h,C0h,C0h,00h");
    asm("Char047: DB 06h,0Ch,18h,30h,60h,C0h,80h,00h");
    asm("Char048: DB 7Ch,CEh,DEh,F6h,E6h,C6h,7Ch,00h");
    asm("Char049: DB 30h,70h,30h,30h,30h,30h,FCh,00h");
    asm("Char050: DB 7Ch,C6h,06h,7Ch,C0h,C0h,FEh,00h");
    asm("Char051: DB FCh,06h,06h,3Ch,06h,06h,FCh,00h");
    asm("Char052: DB 0Ch,CCh,CCh,CCh,FEh,0Ch,0Ch,00h");
    asm("Char053: DB FEh,C0h,FCh,06h,06h,C6h,7Ch,00h");
    asm("Char054: DB 7Ch,C0h,C0h,FCh,C6h,C6h,7Ch,00h");
    asm("Char055: DB FEh,06h,06h,0Ch,18h,30h,30h,00h");
    asm("Char056: DB 7Ch,C6h,C6h,7Ch,C6h,C6h,7Ch,00h");
    asm("Char057: DB 7Ch,C6h,C6h,7Eh,06h,06h,7Ch,00h");
    asm("Char058: DB 00h,C0h,C0h,00h,00h,C0h,C0h,00h");
    asm("Char059: DB 00h,60h,60h,00h,00h,60h,60h,C0h");
    asm("Char060: DB 18h,30h,60h,C0h,60h,30h,18h,00h");
    asm("Char061: DB 00h,00h,FCh,00h,FCh,00h,00h,00h");
    asm("Char062: DB C0h,60h,30h,18h,30h,60h,C0h,00h");
    asm("Char063: DB 78h,CCh,18h,30h,30h,00h,30h,00h");
    asm("Char064: DB 7Ch,C6h,DEh,DEh,DEh,C0h,7Eh,00h");
    asm("Char065: DB 38h,6Ch,C6h,C6h,FEh,C6h,C6h,00h");
    asm("Char066: DB FCh,C6h,C6h,FCh,C6h,C6h,FCh,00h");
    asm("Char067: DB 7Ch,C6h,C0h,C0h,C0h,C6h,7Ch,00h");
    asm("Char068: DB F8h,CCh,C6h,C6h,C6h,CCh,F8h,00h");
    asm("Char069: DB FEh,C0h,C0h,F8h,C0h,C0h,FEh,00h");
    asm("Char070: DB FEh,C0h,C0h,F8h,C0h,C0h,C0h,00h");
    asm("Char071: DB 7Ch,C6h,C0h,C0h,CEh,C6h,7Ch,00h");
    asm("Char072: DB C6h,C6h,C6h,FEh,C6h,C6h,C6h,00h");
    asm("Char073: DB 7Eh,18h,18h,18h,18h,18h,7Eh,00h");
    asm("Char074: DB 06h,06h,06h,06h,06h,C6h,7Ch,00h");
    asm("Char075: DB C6h,CCh,D8h,F0h,D8h,CCh,C6h,00h");
    asm("Char076: DB C0h,C0h,C0h,C0h,C0h,C0h,FEh,00h");
    asm("Char077: DB C6h,EEh,FEh,FEh,D6h,C6h,C6h,00h");
    asm("Char078: DB C6h,E6h,F6h,DEh,CEh,C6h,C6h,00h");
    asm("Char079: DB 7Ch,C6h,C6h,C6h,C6h,C6h,7Ch,00h");
    asm("Char080: DB FCh,C6h,C6h,FCh,C0h,C0h,C0h,00h");
    asm("Char081: DB 7Ch,C6h,C6h,C6h,D6h,DEh,7Ch,06h");
    asm("Char082: DB FCh,C6h,C6h,FCh,D8h,CCh,C6h,00h");
    asm("Char083: DB 7Ch,C6h,C0h,7Ch,06h,C6h,7Ch,00h");
    asm("Char084: DB FFh,18h,18h,18h,18h,18h,18h,00h");
    asm("Char085: DB C6h,C6h,C6h,C6h,C6h,C6h,FEh,00h");
    asm("Char086: DB C6h,C6h,C6h,C6h,C6h,7Ch,38h,00h");
    asm("Char087: DB C6h,C6h,C6h,C6h,D6h,FEh,6Ch,00h");
    asm("Char088: DB C6h,C6h,6Ch,38h,6Ch,C6h,C6h,00h");
    asm("Char089: DB C6h,C6h,C6h,7Ch,18h,30h,E0h,00h");
    asm("Char090: DB FEh,06h,0Ch,18h,30h,60h,FEh,00h");
    asm("Char091: DB F0h,C0h,C0h,C0h,C0h,C0h,F0h,00h");
    asm("Char092: DB C0h,60h,30h,18h,0Ch,06h,02h,00h");
    asm("Char093: DB F0h,30h,30h,30h,30h,30h,F0h,00h");
    asm("Char094: DB 10h,38h,6Ch,C6h,00h,00h,00h,00h");
    asm("Char095: DB 00h,00h,00h,00h,00h,00h,00h,FFh");
    asm("Char096: DB C0h,C0h,60h,00h,00h,00h,00h,00h");
    asm("Char097: DB 00h,00h,7Ch,06h,7Eh,C6h,7Eh,00h");
    asm("Char098: DB C0h,C0h,C0h,FCh,C6h,C6h,FCh,00h");
    asm("Char099: DB 00h,00h,7Ch,C6h,C0h,C6h,7Ch,00h");
    asm("Char100: DB 06h,06h,06h,7Eh,C6h,C6h,7Eh,00h");
    asm("Char101: DB 00h,00h,7Ch,C6h,FEh,C0h,7Ch,00h");
    asm("Char102: DB 1Ch,36h,30h,78h,30h,30h,78h,00h");
    asm("Char103: DB 00h,00h,7Eh,C6h,C6h,7Eh,06h,FCh");
    asm("Char104: DB C0h,C0h,FCh,C6h,C6h,C6h,C6h,00h");
    asm("Char105: DB 60h,00h,E0h,60h,60h,60h,F0h,00h");
    asm("Char106: DB 06h,00h,06h,06h,06h,06h,C6h,7Ch");
    asm("Char107: DB C0h,C0h,CCh,D8h,F8h,CCh,C6h,00h");
    asm("Char108: DB E0h,60h,60h,60h,60h,60h,F0h,00h");
    asm("Char109: DB 00h,00h,CCh,FEh,FEh,D6h,D6h,00h");
    asm("Char110: DB 00h,00h,FCh,C6h,C6h,C6h,C6h,00h");
    asm("Char111: DB 00h,00h,7Ch,C6h,C6h,C6h,7Ch,00h");
    asm("Char112: DB 00h,00h,FCh,C6h,C6h,FCh,C0h,C0h");
    asm("Char113: DB 00h,00h,7Eh,C6h,C6h,7Eh,06h,06h");
    asm("Char114: DB 00h,00h,FCh,C6h,C0h,C0h,C0h,00h");
    asm("Char115: DB 00h,00h,7Eh,C0h,7Ch,06h,FCh,00h");
    asm("Char116: DB 30h,30h,FCh,30h,30h,30h,1Ch,00h");
    asm("Char117: DB 00h,00h,C6h,C6h,C6h,C6h,7Eh,00h");
    asm("Char118: DB 00h,00h,C6h,C6h,C6h,7Ch,38h,00h");
    asm("Char119: DB 00h,00h,C6h,C6h,D6h,FEh,6Ch,00h");
    asm("Char120: DB 00h,00h,C6h,6Ch,38h,6Ch,C6h,00h");
    asm("Char121: DB 00h,00h,C6h,C6h,C6h,7Eh,06h,FCh");
    asm("Char122: DB 00h,00h,FEh,0Ch,38h,60h,FEh,00h");
    asm("Char123: DB 1Ch,30h,30h,E0h,30h,30h,1Ch,00h");
    asm("Char124: DB C0h,C0h,C0h,00h,C0h,C0h,C0h,00h");
    asm("Char125: DB E0h,30h,30h,1Ch,30h,30h,E0h,00h");
    asm("Char126: DB 76h,DCh,00h,00h,00h,00h,00h,00h");
    asm("Char127: DB 00h,10h,38h,6Ch,C6h,C6h,FEh,00h");
    asm("Char128: DB 7Ch,C6h,C0h,C0h,C0h,D6h,7Ch,30h");
    asm("Char129: DB C6h,00h,C6h,C6h,C6h,C6h,7Eh,00h");
    asm("Char130: DB 0Eh,00h,7Ch,C6h,FEh,C0h,7Ch,00h");
    asm("Char131: DB 7Eh,81h,3Ch,06h,7Eh,C6h,7Eh,00h");
    asm("Char132: DB 66h,00h,7Ch,06h,7Eh,C6h,7Eh,00h");
    asm("Char133: DB E0h,00h,7Ch,06h,7Eh,C6h,7Eh,00h");
    asm("Char134: DB 18h,18h,7Ch,06h,7Eh,C6h,7Eh,00h");
    asm("Char135: DB 00h,00h,7Ch,C6h,C0h,D6h,7Ch,30h");
    asm("Char136: DB 7Eh,81h,7Ch,C6h,FEh,C0h,7Ch,00h");
    asm("Char137: DB 66h,00h,7Ch,C6h,FEh,C0h,7Ch,00h");
    asm("Char138: DB E0h,00h,7Ch,C6h,FEh,C0h,7Ch,00h");
    asm("Char139: DB 66h,00h,38h,18h,18h,18h,3Ch,00h");
    asm("Char140: DB 7Ch,82h,38h,18h,18h,18h,3Ch,00h");
    asm("Char141: DB 70h,00h,38h,18h,18h,18h,3Ch,00h");
    asm("Char142: DB C6h,10h,7Ch,C6h,FEh,C6h,C6h,00h");
    asm("Char143: DB 38h,38h,00h,7Ch,C6h,FEh,C6h,00h");
    asm("Char144: DB 0Eh,00h,FEh,C0h,F8h,C0h,FEh,00h");
    asm("Char145: DB 00h,00h,7Fh,0Ch,7Fh,CCh,7Fh,00h");
    asm("Char146: DB 3Fh,6Ch,CCh,FFh,CCh,CCh,CFh,00h");
    asm("Char147: DB 7Ch,82h,7Ch,C6h,C6h,C6h,7Ch,00h");
    asm("Char148: DB 66h,00h,7Ch,C6h,C6h,C6h,7Ch,00h");
    asm("Char149: DB E0h,00h,7Ch,C6h,C6h,C6h,7Ch,00h");
    asm("Char150: DB 7Ch,82h,00h,C6h,C6h,C6h,7Eh,00h");
    asm("Char151: DB E0h,00h,C6h,C6h,C6h,C6h,7Eh,00h");
    asm("Char152: DB 66h,00h,66h,66h,66h,3Eh,06h,7Ch");
    asm("Char153: DB C6h,7Ch,C6h,C6h,C6h,C6h,7Ch,00h");
    asm("Char154: DB C6h,00h,C6h,C6h,C6h,C6h,FEh,00h");
    asm("Char155: DB 18h,18h,7Eh,D8h,D8h,D8h,7Eh,18h");
    asm("Char156: DB 38h,6Ch,60h,F0h,60h,66h,FCh,00h");
    asm("Char157: DB 66h,66h,3Ch,18h,7Eh,18h,7Eh,18h");
    asm("Char158: DB F8h,CCh,CCh,FAh,C6h,CFh,C6h,C3h");
    asm("Char159: DB 0Eh,1Bh,18h,3Ch,18h,18h,D8h,70h");
    asm("Char160: DB 0Eh,00h,7Ch,06h,7Eh,C6h,7Eh,00h");
    asm("Char161: DB 1Ch,00h,38h,18h,18h,18h,3Ch,00h");
    asm("Char162: DB 0Eh,00h,7Ch,C6h,C6h,C6h,7Ch,00h");
    asm("Char163: DB 0Eh,00h,C6h,C6h,C6h,C6h,7Eh,00h");
    asm("Char164: DB 00h,FEh,00h,FCh,C6h,C6h,C6h,00h");
    asm("Char165: DB FEh,00h,C6h,E6h,F6h,DEh,CEh,00h");
    asm("Char166: DB 3Ch,6Ch,6Ch,3Eh,00h,7Eh,00h,00h");
    asm("Char167: DB 3Ch,66h,66h,3Ch,00h,7Eh,00h,00h");
    asm("Char168: DB 18h,00h,18h,18h,30h,66h,3Ch,00h");
    asm("Char169: DB 00h,00h,00h,FCh,C0h,C0h,00h,00h");
    asm("Char170: DB 00h,00h,00h,FCh,0Ch,0Ch,00h,00h");
    asm("Char171: DB C6h,CCh,D8h,3Fh,63h,CFh,8Ch,0Fh");
    asm("Char172: DB C3h,C6h,CCh,DBh,37h,6Dh,CFh,03h");
    asm("Char173: DB 18h,00h,18h,18h,18h,18h,18h,00h");
    asm("Char174: DB 00h,33h,66h,CCh,66h,33h,00h,00h");
    asm("Char175: DB 00h,CCh,66h,33h,66h,CCh,00h,00h");
    asm("Char176: DB 22h,88h,22h,88h,22h,88h,22h,88h");
    asm("Char177: DB 55h,AAh,55h,AAh,55h,AAh,55h,AAh");
    asm("Char178: DB DDh,77h,DDh,77h,DDh,77h,DDh,77h");
    asm("Char179: DB 18h,18h,18h,18h,18h,18h,18h,18h");
    asm("Char180: DB 18h,18h,18h,18h,F8h,18h,18h,18h");
    asm("Char181: DB 18h,18h,F8h,18h,F8h,18h,18h,18h");
    asm("Char182: DB 36h,36h,36h,36h,F6h,36h,36h,36h");
    asm("Char183: DB 00h,00h,00h,00h,FEh,36h,36h,36h");
    asm("Char184: DB 00h,00h,F8h,18h,F8h,18h,18h,18h");
    asm("Char185: DB 36h,36h,F6h,06h,F6h,36h,36h,36h");
    asm("Char186: DB 36h,36h,36h,36h,36h,36h,36h,36h");
    asm("Char187: DB 00h,00h,FEh,06h,F6h,36h,36h,36h");
    asm("Char188: DB 36h,36h,F6h,06h,FEh,00h,00h,00h");
    asm("Char189: DB 36h,36h,36h,36h,FEh,00h,00h,00h");
    asm("Char190: DB 18h,18h,F8h,18h,F8h,00h,00h,00h");
    asm("Char191: DB 00h,00h,00h,00h,F8h,18h,18h,18h");
    asm("Char192: DB 18h,18h,18h,18h,1Fh,00h,00h,00h");
    asm("Char193: DB 18h,18h,18h,18h,FFh,00h,00h,00h");
    asm("Char194: DB 00h,00h,00h,00h,FFh,18h,18h,18h");
    asm("Char195: DB 18h,18h,18h,18h,1Fh,18h,18h,18h");
    asm("Char196: DB 00h,00h,00h,00h,FFh,00h,00h,00h");
    asm("Char197: DB 18h,18h,18h,18h,FFh,18h,18h,18h");
    asm("Char198: DB 18h,18h,1Fh,18h,1Fh,18h,18h,18h");
    asm("Char199: DB 36h,36h,36h,36h,37h,36h,36h,36h");
    asm("Char200: DB 36h,36h,37h,30h,3Fh,00h,00h,00h");
    asm("Char201: DB 00h,00h,3Fh,30h,37h,36h,36h,36h");
    asm("Char202: DB 36h,36h,F7h,00h,FFh,00h,00h,00h");
    asm("Char203: DB 00h,00h,FFh,00h,F7h,36h,36h,36h");
    asm("Char204: DB 36h,36h,37h,30h,37h,36h,36h,36h");
    asm("Char205: DB 00h,00h,FFh,00h,FFh,00h,00h,00h");
    asm("Char206: DB 36h,36h,F7h,00h,F7h,36h,36h,36h");
    asm("Char207: DB 18h,18h,FFh,00h,FFh,00h,00h,00h");
    asm("Char208: DB 36h,36h,36h,36h,FFh,00h,00h,00h");
    asm("Char209: DB 00h,00h,FFh,00h,FFh,18h,18h,18h");
    asm("Char210: DB 00h,00h,00h,00h,FFh,36h,36h,36h");
    asm("Char211: DB 36h,36h,36h,36h,3Fh,00h,00h,00h");
    asm("Char212: DB 18h,18h,1Fh,18h,1Fh,00h,00h,00h");
    asm("Char213: DB 00h,00h,1Fh,18h,1Fh,18h,18h,18h");
    asm("Char214: DB 00h,00h,00h,00h,3Fh,36h,36h,36h");
    asm("Char215: DB 36h,36h,36h,36h,FFh,36h,36h,36h");
    asm("Char216: DB 18h,18h,FFh,18h,FFh,18h,18h,18h");
    asm("Char217: DB 18h,18h,18h,18h,F8h,00h,00h,00h");
    asm("Char218: DB 00h,00h,00h,00h,1Fh,18h,18h,18h");
    asm("Char219: DB FFh,FFh,FFh,FFh,FFh,FFh,FFh,FFh");
    asm("Char220: DB 00h,00h,00h,00h,FFh,FFh,FFh,FFh");
    asm("Char221: DB F0h,F0h,F0h,F0h,F0h,F0h,F0h,F0h");
    asm("Char222: DB 0Fh,0Fh,0Fh,0Fh,0Fh,0Fh,0Fh,0Fh");
    asm("Char223: DB FFh,FFh,FFh,FFh,00h,00h,00h,00h");
    asm("Char224: DB 00h,00h,76h,DCh,C8h,DCh,76h,00h");
    asm("Char225: DB 38h,6Ch,6Ch,78h,6Ch,66h,6Ch,60h");
    asm("Char226: DB 00h,FEh,C6h,C0h,C0h,C0h,C0h,00h");
    asm("Char227: DB 00h,00h,FEh,6Ch,6Ch,6Ch,6Ch,00h");
    asm("Char228: DB FEh,60h,30h,18h,30h,60h,FEh,00h");
    asm("Char229: DB 00h,00h,7Eh,D8h,D8h,D8h,70h,00h");
    asm("Char230: DB 00h,66h,66h,66h,66h,7Ch,60h,C0h");
    asm("Char231: DB 00h,76h,DCh,18h,18h,18h,18h,00h");
    asm("Char232: DB 7Eh,18h,3Ch,66h,66h,3Ch,18h,7Eh");
    asm("Char233: DB 3Ch,66h,C3h,FFh,C3h,66h,3Ch,00h");
    asm("Char234: DB 3Ch,66h,C3h,C3h,66h,66h,E7h,00h");
    asm("Char235: DB 0Eh,18h,0Ch,7Eh,C6h,C6h,7Ch,00h");
    asm("Char236: DB 00h,00h,7Eh,DBh,DBh,7Eh,00h,00h");
    asm("Char237: DB 06h,0Ch,7Eh,DBh,DBh,7Eh,60h,C0h");
    asm("Char238: DB 38h,60h,C0h,F8h,C0h,60h,38h,00h");
    asm("Char239: DB 78h,CCh,CCh,CCh,CCh,CCh,CCh,00h");
    asm("Char240: DB 00h,7Eh,00h,7Eh,00h,7Eh,00h,00h");
    asm("Char241: DB 18h,18h,7Eh,18h,18h,00h,7Eh,00h");
    asm("Char242: DB 60h,30h,18h,30h,60h,00h,FCh,00h");
    asm("Char243: DB 18h,30h,60h,30h,18h,00h,FCh,00h");
    asm("Char244: DB 0Eh,1Bh,1Bh,18h,18h,18h,18h,18h");
    asm("Char245: DB 18h,18h,18h,18h,18h,D8h,D8h,70h");
    asm("Char246: DB 18h,18h,00h,7Eh,00h,18h,18h,00h");
    asm("Char247: DB 00h,76h,DCh,00h,76h,DCh,00h,00h");
    asm("Char248: DB 38h,6Ch,6Ch,38h,00h,00h,00h,00h");
    asm("Char249: DB 00h,00h,00h,18h,18h,00h,00h,00h");
    asm("Char250: DB 00h,00h,00h,00h,18h,00h,00h,00h");
    asm("Char251: DB 0Fh,0Ch,0Ch,0Ch,ECh,6Ch,3Ch,1Ch");
    asm("Char252: DB 78h,6Ch,6Ch,6Ch,6Ch,00h,00h,00h");
    asm("Char253: DB 7Ch,0Ch,7Ch,60h,7Ch,00h,00h,00h");
    asm("Char254: DB 00h,00h,3Ch,3Ch,3Ch,3Ch,00h,00h");
    asm("Char255: DB 00h,10h,00h,00h,00h,00h,00h,00h");
}

void printSmall8bpp(const char* string, uint16_t x, uint8_t y)
{
    asm("LD DE,(IX+9)");
    asm("LD (posX),DE");
    asm("LD A,(IX+12)");
    asm("LD (posY),A");
    asm("LD HL,(IX+6)");
    asm("CALL drawStringSmall");
}

uint8_t isKeyPressed(uint16_t key)
{
    uint8_t* addr = (uint8_t*)(0xF50000 + (uint8_t)(key >> 8));
    uint8_t mask = key & 0xFF;
    return ((*addr)&mask) ? 1 : 0;
}

void scanKeys()
{
    asm("ld hl,0F50000h");
    asm("ld (hl),3");      // Set Continuous Scan mode
//    asm("ld (hl),2");      // Set Single Scan mode
 /*   asm("xor a,a");
    asm("scan_wait:");
    asm("cp a,(hl)");      // Wait for Idle mode
    asm("jr nz,scan_wait");*/
}